## This contains various QP solvers that solve the convex subproblem of the sqp
## Author : Avadesh Meduri & Armand Jordana
## Date : 21/03/2022

import importlib.util
import time

import numpy as np
import osqp
import proxsuite
from crocoddyl import SolverAbstract
from scipy import sparse

from .py_osqp import CustomOSQP
from .stagewise_qp_kkt import StagewiseQPKKT

HPIPM_PYTHON_FOUND = importlib.util.find_spec("hpipm_python")
if HPIPM_PYTHON_FOUND:
    import hpipm_python


class QPSolvers(SolverAbstract, CustomOSQP, StagewiseQPKKT):
    def __init__(self, shootingProblem, method, verboseQP=True):
        SolverAbstract.__init__(self, shootingProblem)

        assert (
            method == "ProxQP"
            or method == "OSQP"
            or method == "CustomOSQP"
            or method == "StagewiseQPKKT"
            or method == "HPIPM_DENSE"
            or method == "HPIPM_OCP"
        )
        self.method = method
        if method == "CustomOSQP":
            CustomOSQP.__init__(self)
        if method == "StagewiseQPKKT":
            StagewiseQPKKT.__init__(self)
        self.allocateDataQP()
        self.max_qp_iters = 1000
        self.initialize = True
        self.verboseQP = verboseQP
        self.eps_abs = 1e-4
        self.eps_rel = 0.0
        self.OSQP_scaling = False

        self.BENCHMARK = True
        self.DEBUG = False

        if self.verboseQP:
            print("USING " + str(method))

    def models(self):
        mod = [m for m in self.problem.runningModels]
        mod += [self.problem.terminalModel]
        return mod

    def calc(self, recalc=True):
        # compute cost and derivatives at deterministic nonlinear trajectory
        if recalc:
            self.problem.calc(self.xs, self.us)
            self.problem.calcDiff(self.xs, self.us)
        self.cost = 0
        self.constraint_norm = 0

        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            # Inequality
            self.constraint_norm += np.linalg.norm(
                np.clip(model.g_lb - data.g, 0, np.inf), 1
            )
            self.constraint_norm += np.linalg.norm(
                np.clip(data.g - model.g_ub, 0, np.inf), 1
            )
            # Equality
            # self.constraint_norm += np.linalg.norm(data.h - model.h_ub, 0, np.inf), 1)

        self.constraint_norm += np.linalg.norm(
            np.clip(
                self.problem.terminalModel.g_lb - self.problem.terminalData.g, 0, np.inf
            ),
            1,
        )
        self.constraint_norm += np.linalg.norm(
            np.clip(
                self.problem.terminalData.g - self.problem.terminalModel.g_ub, 0, np.inf
            ),
            1,
        )

        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            self.gap[t] = model.state.diff(
                self.xs[t + 1].copy(), data.xnext.copy()
            )  # gaps
            self.cost += data.cost

        self.gap_norm = sum(np.linalg.norm(self.gap.copy(), 1, axis=1))
        self.cost += self.problem.terminalData.cost
        self.gap = self.gap.copy()

    def computeDirectionFullQP(self):
        self.n_vars = self.problem.T * (self.ndx + self.nu)

        P = np.zeros(
            (
                self.problem.T * (self.ndx + self.nu),
                self.problem.T * (self.ndx + self.nu),
            )
        )
        q = np.zeros(self.problem.T * (self.ndx + self.nu))

        Asize = self.problem.T * (self.ndx + self.nu)
        A = np.zeros((self.problem.T * self.ndx, Asize))
        B = np.zeros(self.problem.T * self.ndx)

        NNZ_block_P = 0
        NNZ_block_A = 0
        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            index_u = self.problem.T * self.ndx + t * self.nu
            if t >= 1:
                index_x = (t - 1) * self.ndx
                P[index_x : index_x + self.ndx, index_x : index_x + self.ndx] = (
                    data.Lxx.copy()
                )
                NNZ_block_P += np.size(data.Lxx)
                P[index_x : index_x + self.ndx, index_u : index_u + self.nu] = (
                    data.Lxu.copy()
                )
                NNZ_block_P += np.size(data.Lxu)
                P[index_u : index_u + self.nu, index_x : index_x + self.ndx] = (
                    data.Lxu.T.copy()
                )
                NNZ_block_P += np.size(data.Lxu.T)
                q[index_x : index_x + self.ndx] = data.Lx.copy()

            P[index_u : index_u + self.nu, index_u : index_u + self.nu] = (
                data.Luu.copy()
            )
            NNZ_block_P += np.size(data.Luu)
            q[index_u : index_u + self.nu] = data.Lu.copy()

            A[
                t * self.ndx : (t + 1) * self.ndx, index_u : index_u + self.nu
            ] = -data.Fu.copy()
            NNZ_block_A += np.size(data.Fu)
            A[t * self.ndx : (t + 1) * self.ndx, t * self.ndx : (t + 1) * self.ndx] = (
                np.eye(self.ndx)
            )
            NNZ_block_A += self.ndx

            if t >= 1:
                A[
                    t * self.ndx : (t + 1) * self.ndx, (t - 1) * self.ndx : t * self.ndx
                ] = -data.Fx.copy()
                NNZ_block_A += np.size(data.Fx)
            B[t * self.ndx : (t + 1) * self.ndx] = self.gap[t].copy()

        P[
            (self.problem.T - 1) * self.ndx : self.problem.T * self.ndx,
            self.problem.T * self.ndx - self.ndx : self.problem.T * self.ndx,
        ] = self.problem.terminalData.Lxx.copy()
        NNZ_block_P += np.size(self.problem.terminalData.Lxx)
        q[(self.problem.T - 1) * self.ndx : self.problem.T * self.ndx] = (
            self.problem.terminalData.Lx.copy()
        )

        n = self.problem.T * (self.ndx + self.nu)
        self.n_eq = self.problem.T * self.ndx

        self.n_in = sum([len(self.y[i]) for i in range(len(self.y))])

        C = np.zeros((self.n_in, self.problem.T * (self.ndx + self.nu)))
        l = np.zeros(self.n_in)  # noqa: E741
        u = np.zeros(self.n_in)

        nin_count = 0
        index_x = self.problem.T * self.ndx
        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            if model.ng == 0:
                continue
            l[nin_count : nin_count + model.ng] = model.g_lb - data.g
            u[nin_count : nin_count + model.ng] = model.g_ub - data.g
            if t > 0:
                C[
                    nin_count : nin_count + model.ng, (t - 1) * self.ndx : t * self.ndx
                ] = data.Gx
                NNZ_block_A += np.size(data.Gx)
            C[
                nin_count : nin_count + model.ng,
                index_x + t * self.nu : index_x + (t + 1) * self.nu,
            ] = data.Gu
            NNZ_block_A += np.size(data.Gu)
            nin_count += model.ng

        model = self.problem.terminalModel
        data = self.problem.terminalData
        if model.ng != 0:
            l[nin_count : nin_count + model.ng] = model.g_lb - data.g
            u[nin_count : nin_count + model.ng] = model.g_ub - data.g
            C[
                nin_count : nin_count + model.ng,
                (self.problem.T - 1) * self.ndx : self.problem.T * self.ndx,
            ] = data.Gx
            NNZ_block_A += np.size(data.Gx)
            nin_count += model.ng

        # If we are benchmarking, setup the problem using QPSolvers (S. Caron)
        # in order to get the unified convergence metrics (prim_res, dual_res, dual_gap)
        if self.BENCHMARK:
            # We follow the convention of QPSolvers package, i.e.
            # min_x (1/2) x^T P x + q^T x
            #   s.t. Ax  = b
            #        Gx <= h
            # see here https://github.com/qpsolvers/qpsolvers/blob/main/qpsolvers/problem.py
            P_qp = P
            q_qp = q
            A_qp = A
            b_qp = B
            G_qp = np.vstack([C, -C])
            h_qp = np.hstack([u, -l])
            # prob_qp = Problem(P_qp, q_qp, G_qp, h_qp, A_qp, b_qp) # no box constraints
            # sol_qp = Solution(prob_qp)

        if self.method == "ProxQP":
            qp = proxsuite.proxqp.sparse.QP(n, self.n_eq, self.n_in)
            qp.settings.eps_abs = self.eps_abs
            qp.settings.eps_rel = self.eps_rel
            qp.settings.max_iter = self.max_qp_iters
            qp.settings.initial_guess = proxsuite.proxqp.NO_INITIAL_GUESS
            qp.init(P, q, A, B, C, l, u)
            t1 = time.time()
            qp.solve()
            self.qp_time = time.time() - t1
            self.qp_iters = qp.results.info.iter
            self.found_qp_sol = (
                qp.results.info.status == proxsuite.proxqp.QPSolverOutput.PROXQP_SOLVED
            )
            # Get solution
            res = qp.results.x
            self.z_k = qp.results.z
            self.y_k = qp.results.y
            # If we are benchmarking the QP, check convergence using dual and primal residuals
            # Check here : https://scaron.info/blog/optimality-conditions-and-numerical-tolerances-in-qp-solvers.html
            if self.BENCHMARK:
                self.norm_dual = np.max(
                    np.abs(P @ res + q + A.T @ self.y_k + C.T @ self.z_k)
                )
                self.norm_primal = np.max(np.abs(A @ res - B))
                if C.shape[0] > 0:
                    res_lb = np.max(np.abs(np.clip(l - C @ res, 0, np.inf)))
                    res_ub = np.max(np.abs(np.clip(C @ res - u, 0, np.inf)))
                    self.norm_primal = np.maximum(res_lb, self.norm_primal)
                    self.norm_primal = np.maximum(res_ub, self.norm_primal)
                print(f"- Primal residual: {self.norm_primal:.1e}")
                print(f"- Dual residual: {self.norm_dual:.1e}")

            # Fill out Lagrange mutlipliers for SQP
            for t in range(self.problem.T):
                self.lag_mul[t + 1] = -qp.results.y[t * self.ndx : (t + 1) * self.ndx]
            nin_count = 0
            for t in range(self.problem.T + 1):
                if t < self.problem.T:
                    model = self.problem.runningModels[t]
                else:
                    model = self.problem.terminalModel
                if model.ng == 0:
                    continue
                self.y[t] = qp.results.z[nin_count : nin_count + model.ng]
                nin_count += model.ng

        elif self.method == "OSQP":
            Aeq = sparse.csc_matrix(A)
            Aineq = sparse.csc_matrix(C)
            Aosqp = sparse.vstack([Aeq, Aineq])
            losqp = np.hstack([B, l])
            uosqp = np.hstack([B, u])
            P = sparse.csc_matrix(P)
            if self.DEBUG:
                print(
                    "nnz(P) = ",
                    P.count_nonzero(),
                    " out of ",
                    NNZ_block_P,
                    " = ",
                    100 * P.count_nonzero() / NNZ_block_P,
                )
                print(
                    "nnz(A) = ",
                    Aosqp.count_nonzero(),
                    " out of ",
                    NNZ_block_A,
                    " = ",
                    100 * Aosqp.count_nonzero() / NNZ_block_A,
                )
            prob = osqp.OSQP()
            prob.setup(
                P,
                q,
                Aosqp,
                losqp,
                uosqp,
                warm_starting=False,
                scaling=self.OSQP_scaling,
                max_iter=self.max_qp_iters,
                adaptive_rho=True,
                adaptive_rho_interval=25,
                verbose=self.verboseQP,
                eps_rel=self.eps_rel,
                eps_abs=self.eps_abs,
            )

            t1 = time.time()
            tmp = prob.solve()
            self.qp_time = time.time() - t1
            self.qp_iters = tmp.info.iter
            self.found_qp_sol = tmp.info.status_val == osqp.constant("OSQP_SOLVED")
            # Get solution
            FOUND_NONE = False
            if None in tmp.y:
                FOUND_NONE = True
                self.y_k = np.zeros(tmp.y.shape)
            else:
                self.y_k = tmp.y
            if None in tmp.x:
                FOUND_NONE = True
                res = np.zeros(tmp.x.shape)
            else:
                res = tmp.x
            # If we are benchmarking the QP, check convergence using dual and primal residuals
            # Check here : https://scaron.info/blog/optimality-conditions-and-numerical-tolerances-in-qp-solvers.html
            if self.BENCHMARK:
                m = C.shape[0]
                meq = A.shape[0]
                if FOUND_NONE:
                    self.norm_primal = np.inf
                    self.norm_dual = np.inf
                else:
                    self.norm_primal = np.max(np.abs(A @ res - B))
                    self.norm_dual = np.max(
                        np.abs(
                            P @ res
                            + q
                            + A.T @ self.y_k[:meq]
                            + C.T @ self.y_k[meq : meq + m]
                        )
                    )
                if C.shape[0] > 0:
                    if FOUND_NONE:
                        res_lb = np.inf
                        res_ub = np.inf
                    else:
                        res_lb = np.max(np.abs(np.clip(l - C @ res, 0, np.inf)))
                        res_ub = np.max(np.abs(np.clip(C @ res - u, 0, np.inf)))
                    self.norm_primal = np.maximum(res_lb, self.norm_primal)
                    self.norm_primal = np.maximum(res_ub, self.norm_primal)
                print(f"- Primal residual: {self.norm_primal:.1e}")
                print(f"- Dual residual: {self.norm_dual:.1e}")

        elif self.method == "HPIPM_DENSE":
            # Dimensions
            nv = self.problem.T * (self.ndx + self.nu)  # number of variables
            ne = (
                self.n_eq if self.n_eq is not None else 0
            )  # number of equality constraints
            ng = (
                self.n_in if self.n_in is not None else 0
            )  # number of general (inequality) constraints
            nb = 0  # number of box constraints
            dim = hpipm_python.hpipm_dense_qp_dim()
            dim.set("nv", nv)
            dim.set("nb", nb)
            dim.set("ne", ne)
            dim.set("ng", ng)
            # QP setup
            qp = hpipm_python.hpipm_dense_qp(dim)
            qp.set("H", P)
            qp.set("g", q)
            if ne > 0:
                qp.set("A", A)
                qp.set("b", B)
            if ng > 0:
                qp.set("C", C)
                # need to mask out lb or ub if the box constraints are only one-sided
                # we also mask out infinities (and set the now-irrelevant value to
                # zero), since HPIPM doesn't like infinities
                lg = l
                # import pdb ; pdb.set_trace()
                if lg is not None:
                    # Detect infs and replace them by 0s
                    lg_mask = np.isinf(lg)
                    lg[lg_mask] = 0.0
                    qp.set("lg", lg)
                    # De-activate those constraints
                    qp.set("lg_mask", ~lg_mask)
                else:
                    qp.set("lg_mask", np.zeros(ng, dtype=bool))
                ug = u
                if ug is not None:
                    ug_mask = np.isinf(ug)
                    ug[ug_mask] = 0.0
                    qp.set("ug", ug)
                    qp.set("ug_mask", ~ug_mask)
                else:
                    qp.set("ug_mask", np.zeros(ng, dtype=bool))
            qp_sol = hpipm_python.hpipm_dense_qp_sol(dim)
            # set up solver arg
            # see default args here https://github.com/giaf/hpipm/blob/f7c4f502172b8ea5279c8f4d34afe882407526c3/dense_qp/x_dense_qp_ipm.c#L104
            mode = "speed"
            # create and set default arg based on mode
            arg = hpipm_python.hpipm_dense_qp_solver_arg(dim, mode)
            # create and set default arg based on mode
            arg.set("iter_max", self.max_qp_iters)
            arg.set("tol_comp", self.eps_abs)
            arg.set("tol_eq", self.eps_abs)
            arg.set("tol_ineq", self.eps_abs)
            arg.set("tol_stat", self.eps_abs)
            solver = hpipm_python.hpipm_dense_qp_solver(dim, arg)
            t1 = time.time()
            solver.solve(qp, qp_sol)
            self.qp_time = time.time() - t1
            self.qp_iters = solver.get("iter")
            self.found_qp_sol = solver.get("status") == 0
            # Get solution
            res = qp_sol.get("v").flatten()
            # TODO: retrieve dual solution for SQP
            self.y_k = np.zeros(self.n_in + self.n_eq)
            self.y_k[: self.n_eq] = (
                -qp_sol.get("pi").flatten() if ne > 0 else np.empty((0,))
            )
            # Lagrange multiplier associated with the box constraint can be determined
            # based on complementarity slackness conditions
            lam_lg = qp_sol.get("lam_lg").flatten()
            lam_ug = qp_sol.get("lam_ug").flatten()
            self.y_k[self.n_eq : self.n_eq + self.n_in] = (
                lam_ug if ng > 0 else np.empty((0,))
            )  # This is probably wrong !!!!!
            # If we are benchmarking the QP, check convergence using dual and primal residuals
            # Check here : https://scaron.info/blog/optimality-conditions-and-numerical-tolerances-in-qp-solvers.html
            if self.BENCHMARK:
                # HPIPM residuals
                self.norm_primal = np.maximum(
                    solver.get("max_res_eq"), solver.get("max_res_ineq")
                )
                self.norm_dual = solver.get("max_res_stat")
                print(f"- Primal residual: {self.norm_primal:.1e}")
                print(f"- Dual residual: {self.norm_dual:.1e}")
                # Optionally check that the residual norm returned by HPIPM matches
                # - hand-calculated residual of full QP
                if self.DEBUG:
                    meq = A_qp.shape[0]
                    # Lagrange multipliers for inequality constraint Gx <= h ( a.k.a. "Cx <= ub" and "-Cx <= -lb" )
                    z_qp = np.hstack([lam_ug, lam_lg])
                    self.norm_primal2 = max(
                        [
                            0.0,
                            np.max(G_qp.dot(res) - h_qp) if G_qp.shape[0] > 0 else 0.0,
                            np.linalg.norm(A_qp.dot(res) - b_qp, np.inf)
                            if A_qp is not None
                            else 0.0,
                        ]
                    )
                    self.norm_dual2 = np.linalg.norm(
                        P_qp.dot(res)
                        + q_qp
                        + G_qp.T.dot(z_qp)
                        + A_qp.T.dot(self.y_k[:meq]),
                        np.inf,
                    )
                    if solver.get("status") == 0:
                        assert np.abs(self.norm_primal - self.norm_primal2) < 1e-12
                        assert np.abs(self.norm_dual - self.norm_dual2) < 1e-12
                    print(f"- Primal residual (full) : {self.norm_primal2:.1e}")
                    print(f"- Dual residual (full) : {self.norm_dual2:.1e}")

                    # HPIPM native callbacks
                    if self.VERBOSE_DEBUG:
                        status = solver.get("status")
                        res_stat = solver.get("max_res_stat")
                        res_eq = solver.get("max_res_eq")
                        res_ineq = solver.get("max_res_ineq")
                        res_comp = solver.get("max_res_comp")
                        iters = solver.get("iter")
                        stat = solver.get("stat")
                        print("\nsolver statistics:\n")
                        print("ipm return = {0:1d}\n".format(status))
                        print("ipm max res stat = {:e}\n".format(res_stat))
                        print("ipm max res eq   = {:e}\n".format(res_eq))
                        print("ipm max res ineq = {:e}\n".format(res_ineq))
                        print("ipm max res comp = {:e}\n".format(res_comp))
                        print("ipm iter = {0:1d}\n".format(iters))
                        print("stat =")
                        print(
                            "\titer\talpha_aff\tmu_aff\t\tsigma\t\talpha_prim\talpha_dual\tmu\t\tres_stat\tres_eq\t\tres_ineq\tres_comp"
                        )
                        for ii in range(iters + 1):
                            print(
                                "\t{:d}\t{:e}\t{:e}\t{:e}\t{:e}\t{:e}\t{:e}\t{:e}\t{:e}\t{:e}\t{:e}".format(
                                    ii,
                                    stat[ii][0],
                                    stat[ii][1],
                                    stat[ii][2],
                                    stat[ii][3],
                                    stat[ii][4],
                                    stat[ii][5],
                                    stat[ii][6],
                                    stat[ii][7],
                                    stat[ii][8],
                                    stat[ii][9],
                                )
                            )
                        print("")

        elif self.method == "HPIPM_OCP":
            # see here https://docs.acados.org/python_interface/index.html?highlight=nbxe#acados-ocp-solver-interface-acadosocp-and-acadosocpsolver
            # Lagrange multipliers only available here https://github.com/giaf/hpipm/pull/159
            # Dimensions
            ne = (
                self.n_eq if self.n_eq is not None else 0
            )  # number of equality constraints
            ng = (
                self.n_in if self.n_in is not None else 0
            )  # number of general (inequality) constraints
            N = self.problem.T
            dim = hpipm_python.hpipm_ocp_qp_dim(N)
            dim.set("nx", self.ndx, 0, N)
            dim.set("nu", self.nu, 0, N - 1)
            for t in range(N):
                dim.set("ng", self.problem.runningModels[t].ng, t)
            dim.set("ng", self.problem.terminalModel.ng, N)
            # QP setup
            qp = hpipm_python.hpipm_ocp_qp(dim)
            # Fill out OCP running nodes
            for t, (model, data) in enumerate(
                zip(self.problem.runningModels, self.problem.runningDatas)
            ):
                # Dynamics
                if t >= 1:
                    qp.set("A", data.Fx, t)
                qp.set("B", data.Fu, t)
                qp.set("b", self.gap[t], t)
                # Cost
                qp.set("R", data.Luu, t)
                qp.set("r", data.Lu.T, t)
                if t >= 1:
                    qp.set("S", data.Lxu.T, t)
                    qp.set("Q", data.Lxx, t)
                    qp.set("q", data.Lx.T, t)
                # Constraints
                if t >= 1:
                    qp.set("C", data.Gx, t)
                qp.set("D", data.Gu, t)
                # need to mask out lb or ub if the box constraints are only one-sided
                # we also mask out infinities (and set the now-irrelevant value to
                # zero), since HPIPM doesn't like infinities
                lg = model.g_lb.copy()
                if lg is not None:
                    # Detect infs and replace them by 0s
                    lg_mask = np.isinf(lg)
                    lg[lg_mask] = 0.0
                    qp.set("lg", lg - data.g, t)
                    # De-activate those constraints
                    qp.set("lg_mask", ~lg_mask, t)
                else:
                    qp.set("lg_mask", np.zeros(ng, dtype=bool))
                ug = model.g_ub.copy()
                if ug is not None:
                    ug_mask = np.isinf(ug)
                    ug[ug_mask] = 0.0
                    qp.set("ug", ug - data.g, t)
                    qp.set("ug_mask", ~ug_mask, t)
                else:
                    qp.set("ug_mask", np.zeros(ng, dtype=bool))

            # Terminal node
            if self.problem.terminalModel.ng > 0:
                lg = self.problem.terminalModel.g_lb.copy()
                if lg is not None:
                    # Detect infs and replace them by 0s
                    lg_mask = np.isinf(lg)
                    lg[lg_mask] = 0.0
                    qp.set("lg", lg - self.problem.terminalData.g, N)
                    # De-activate those constraints
                    qp.set("lg_mask", ~lg_mask, N)
                else:
                    qp.set("lg_mask", np.zeros(ng, dtype=bool))
                ug = self.problem.terminalModel.g_ub.copy()
                if ug is not None:
                    ug_mask = np.isinf(ug)
                    ug[ug_mask] = 0.0
                    qp.set("ug", ug - self.problem.terminalData.g, N)
                    qp.set("ug_mask", ~ug_mask, N)
                else:
                    qp.set("ug_mask", np.zeros(ng, dtype=bool))
                qp.set("C", self.problem.terminalData.Gx, N)
            qp.set("Q", self.problem.terminalData.Lxx, N)
            qp.set("q", self.problem.terminalData.Lx.T, N)
            qp_sol = hpipm_python.hpipm_ocp_qp_sol(dim)
            # set up solver arg
            mode = "speed"
            # create and set default arg based on mode
            arg = hpipm_python.hpipm_ocp_qp_solver_arg(dim, mode)
            # create and set default arg based on mode
            arg.set("iter_max", self.max_qp_iters)
            arg.set("tol_comp", self.eps_abs)
            arg.set("tol_eq", self.eps_abs)
            arg.set("tol_ineq", self.eps_abs)
            arg.set("tol_stat", self.eps_abs)
            solver = hpipm_python.hpipm_ocp_qp_solver(dim, arg)
            t1 = time.time()
            solver.solve(qp, qp_sol)
            self.qp_time = time.time() - t1
            self.qp_iters = solver.get("iter")
            self.found_qp_sol = solver.get("status") == 0
            # Get solution (need to stack the full QP solution for SQP applications)
            res = np.zeros(self.problem.T * (self.ndx + self.nu))
            self.y_k = np.zeros(self.n_in + self.n_eq)
            self.dx[0] = np.zeros(self.ndx)
            self.lag_mul[0] = np.zeros(self.problem.runningModels[0].state.ndx)
            for t in range(self.problem.T):
                # Stacked (full QP) solution
                res[t * self.ndx : (t + 1) * self.ndx] = qp_sol.get(
                    "x", t + 1
                ).flatten()
                index_u = self.problem.T * self.ndx + t * self.nu
                res[index_u : index_u + self.nu] = qp_sol.get("u", t).flatten()
                # Sequential solution (for KKT computation)
                self.du[t] = qp_sol.get("u", t).flatten()
                self.dx[t + 1] = qp_sol.get("x", t + 1).flatten()
                # Lagrange multipliers associated with dynamics constraint
                if self.DEBUG:
                    self.y_k[t * self.ndx : (t + 1) * self.ndx] = -qp_sol.get(
                        "pi", t
                    ).flatten()
                    self.lag_mul[t + 1] = qp_sol.get("pi", t).flatten()
                    # TODO: retrieve dual solution for SQP
                    # Lagrange multiplier associated with the box constraint can be determined
                    # based on complementarity slackness conditions
                    lam_ug = np.zeros(self.n_in)
                    lam_lg = np.zeros(self.n_in)
                    # lam    = np.zeros(self.n_in)
                    nin_count = 0
                    for t in range(self.problem.T + 1):
                        if t < self.problem.T:
                            model = self.problem.runningModels[t]
                            # data = self.problem.runningDatas[t]
                            # tmp_vec = data.Gu @ self.du[t]
                            # if(t > 0):
                            #     tmp_vec += data.Gx @ self.dx[t]
                        else:
                            model = self.problem.terminalModel
                            # data = self.problem.terminalData
                            # tmp_vec = data.Gx @ self.dx[t]
                        if model.ng > 0:
                            lam_ug[nin_count : nin_count + model.ng] = qp_sol.get(
                                "lam_ug", t
                            ).flatten()
                            lam_lg[nin_count : nin_count + model.ng] = qp_sol.get(
                                "lam_lg", t
                            ).flatten()
                            # # "unified" lagrange multipliers : check constraint residual of the solutiontmp_vec
                            # if(np.linalg.norm(tmp_vec - model.g_lb, np.inf) < 1e-6 and np.linalg.norm(qp_sol.get("lam_ug", t)) < 1e-6):
                            #     lam[nin_count: nin_count+model.ng] = qp_sol.get("lam_lg", t).flatten()
                            # elif(np.linalg.norm(tmp_vec - model.g_lb, np.inf) < 1e-6 and np.linalg.norm(qp_sol.get("lam_lg", t)) < 1e-6):
                            #     lam[nin_count: nin_count+model.ng] = qp_sol.get("lam_ug", t).flatten()
                        nin_count += model.ng
                    self.y_k[self.n_eq : self.n_eq + self.n_in] = (
                        lam_ug  # This is probably wrong !!!!! #TODO use 'lam' instead (needs debug)
                    )
            if self.BENCHMARK:
                # HPIPM residuals
                self.norm_primal = np.maximum(
                    solver.get("max_res_eq"), solver.get("max_res_ineq")
                )
                self.norm_dual = solver.get("max_res_stat")
                print(f"- Primal residual: {self.norm_primal:.1e}")
                print(f"- Dual residual: {self.norm_dual:.1e}")
                # Optionally check that the residual norm returned by HPIPM matches
                # 1) hand-calculated residual of full QP
                # 2) KKT criteria (sequential form)
                if self.DEBUG:
                    # 1) Full QP
                    # Lagrange multipliers for inequality constraint Gx <= h ( a.k.a. "Cx <= ub" and "-Cx <= -lb" )
                    meq = A_qp.shape[0]
                    self.norm_primal2 = max(
                        [
                            0.0,
                            max(G_qp @ res - h_qp) if G_qp.shape[0] > 0 else 0.0,
                            max(A_qp @ res - b_qp) if A_qp.shape[0] > 0 else 0.0,
                        ]
                    )
                    self.norm_dual2 = np.linalg.norm(
                        P_qp.dot(res)
                        + q_qp
                        + C.T @ lam_ug
                        - C.T @ lam_lg
                        + A_qp.T.dot(self.y_k[:meq]),
                        np.inf,
                    )
                    if solver.get("status") == 0:
                        assert (
                            np.linalg.norm(self.norm_primal - self.norm_primal2) < 1e-12
                        )
                        assert np.linalg.norm(self.norm_dual - self.norm_dual2) < 1e-12
                    print(
                        f"- [DEBUG] Primal residual (full QP) : {self.norm_primal2:.1e}"
                    )
                    print(f"- [DEBUG] Dual residual (full QP) : {self.norm_dual2:.1e}")

                    # 2) Check sequential KKT residual
                    KKT = 0
                    l1 = 0
                    l2 = 0
                    for t, (model, data) in enumerate(
                        zip(self.problem.runningModels, self.problem.runningDatas)
                    ):
                        Cx, Cu = data.Gx, data.Gu
                        if t == 0:
                            lu = (
                                data.Luu @ self.du[t]
                                + data.Lxu.T @ self.dx[t]
                                + data.Lu
                                + data.Fu.T @ self.lag_mul[t + 1]
                                + Cu.T @ qp_sol.get("lam_ug", t).flatten()
                                - Cu.T @ qp_sol.get("lam_lg", t).flatten()
                            )
                            KKT = max(KKT, max(abs(lu)))
                            continue
                        lx = (
                            data.Lxx @ self.dx[t]
                            + data.Lxu @ self.du[t]
                            + data.Lx
                            + data.Fx.T @ self.lag_mul[t + 1]
                            - self.lag_mul[t]
                            + Cx.T @ qp_sol.get("lam_ug", t).flatten()
                            - Cx.T @ qp_sol.get("lam_lg", t).flatten()
                        )
                        lu = (
                            data.Luu @ self.du[t]
                            + data.Lxu.T @ self.dx[t]
                            + data.Lu
                            + data.Fu.T @ self.lag_mul[t + 1]
                            + Cu.T @ qp_sol.get("lam_ug", t).flatten()
                            - Cu.T @ qp_sol.get("lam_lg", t).flatten()
                        )
                        KKT = max(KKT, max(abs(lx)), max(abs(lu)))
                        if model.ng != 0:
                            l1 = np.max(
                                np.abs(
                                    np.clip(
                                        model.g_lb
                                        - Cx @ self.dx[t]
                                        - Cu @ self.du[t]
                                        - data.g,
                                        0,
                                        np.inf,
                                    )
                                )
                            )
                            l2 = np.max(
                                np.abs(
                                    np.clip(
                                        Cx @ self.dx[t]
                                        + Cu @ self.du[t]
                                        + data.g
                                        - model.g_ub,
                                        0,
                                        np.inf,
                                    )
                                )
                            )
                        l3 = np.max(
                            np.abs(
                                self.dx[t + 1]
                                - data.Fx @ self.dx[t]
                                - data.Fu @ self.du[t]
                                - self.gap[t]
                            )
                        )
                        KKT = max(KKT, l1, l2, l3)
                    model = self.problem.terminalModel
                    data = self.problem.terminalData
                    Cx = data.Gx
                    if model.ng != 0:
                        l1 = np.max(
                            np.abs(
                                np.clip(
                                    model.g_lb - Cx @ self.dx[-1] - data.g, 0, np.inf
                                )
                            )
                        )
                        l2 = np.max(
                            np.abs(
                                np.clip(
                                    Cx @ self.dx[-1] + data.g - model.g_ub, 0, np.inf
                                )
                            )
                        )
                    lx = (
                        self.problem.terminalData.Lxx @ self.dx[-1]
                        + self.problem.terminalData.Lx
                        - self.lag_mul[-1]
                        + Cx.T @ qp_sol.get("lam_ug", N).flatten()
                        - Cx.T @ qp_sol.get("lam_lg", N).flatten()
                    )
                    KKT = max(KKT, l1, l2)
                    KKT = max(KKT, max(abs(lx)))
                    assert (
                        np.linalg.norm(KKT - max(self.norm_primal, self.norm_dual))
                        < 1e-12
                    )
                    print(f"- [DEBUG] KKT residual (sequential form) : {KKT:.1e}")

                    # HPIPM native callbacks
                    if self.DEBUG_VERBOSE:
                        # get solver statistics
                        # 'lam' returned by HPIPM print is equal to ( lam_lg[1], lam_ug[1], ..., lam_lg[N], lam_ug[N] )
                        # qp_sol.print_C_struct()
                        status = solver.get("status")
                        res_stat = solver.get("max_res_stat")
                        res_eq = solver.get("max_res_eq")
                        res_ineq = solver.get("max_res_ineq")
                        res_comp = solver.get("max_res_comp")
                        iters = solver.get("iter")
                        stat = solver.get("stat")
                        print("\nsolver statistics:\n")
                        print("ipm return = {0:1d}\n".format(status))
                        print("ipm max res stat = {:e}\n".format(res_stat))
                        print("ipm max res eq   = {:e}\n".format(res_eq))
                        print("ipm max res ineq = {:e}\n".format(res_ineq))
                        print("ipm max res comp = {:e}\n".format(res_comp))
                        print("ipm iter = {0:1d}\n".format(iters))
                        print("stat =")
                        print(
                            "\titer\talpha_aff\tmu_aff\t\tsigma\t\talpha_prim\talpha_dual\tmu\t\tres_stat\tres_eq\t\tres_ineq\tres_comp"
                        )
                        for ii in range(iters + 1):
                            print(
                                "\t{:d}\t{:e}\t{:e}\t{:e}\t{:e}\t{:e}\t{:e}\t{:e}\t{:e}\t{:e}\t{:e}".format(
                                    ii,
                                    stat[ii][0],
                                    stat[ii][1],
                                    stat[ii][2],
                                    stat[ii][3],
                                    stat[ii][4],
                                    stat[ii][5],
                                    stat[ii][6],
                                    stat[ii][7],
                                    stat[ii][8],
                                    stat[ii][9],
                                )
                            )
                        print("")

        elif self.method == "CustomOSQP":
            Aeq = sparse.csr_matrix(A)
            Aineq = sparse.csr_matrix(C)
            self.Aosqp = sparse.vstack([Aeq, Aineq])

            self.losqp = np.hstack([B, l])
            self.uosqp = np.hstack([B, u])

            self.P = P
            self.q = np.array(q)

            self.x_k = np.zeros(self.n_vars)
            self.z_k = np.zeros(self.n_in + self.n_eq)
            self.y_k = np.zeros(self.n_in + self.n_eq)
            t0 = time.time()
            res = self.optimize_osqp(maxiters=self.max_qp_iters)
            self.qp_time = time.time() - t0
            self.found_qp_sol = self.converged
            self.norm_primal = self.r_prim
            self.norm_dual = self.r_dual

        elif self.method == "StagewiseQPKKT":
            self.A_eq = sparse.csr_matrix(A.copy())
            self.A_in = sparse.csr_matrix(C.copy())
            self.b = B.copy()
            self.lboyd = l.copy()
            self.uboyd = u.copy()

            self.P = P.copy()
            self.q = np.array(q).copy()

            if self.initialize:
                self.xs_vec = np.array(self.xs).flatten()[self.ndx :]
                self.us_vec = np.array(self.us).flatten()
                self.x_k = np.zeros_like(np.hstack((self.xs_vec, self.us_vec)))
                self.z_k = np.zeros(self.n_in)
                self.y_k = np.zeros(self.n_in)

                self.initialize = False

            res = self.optimize_boyd(maxiters=self.max_qp_iters)
            for t in range(self.problem.T):
                self.lag_mul[t + 1] = -self.v_k_1[t * self.ndx : (t + 1) * self.ndx]

            nin_count = 0
            for t in range(self.problem.T + 1):
                if t < self.problem.T:
                    model = self.problem.runningModels[t]
                else:
                    model = self.problem.terminalModel
                if model.ng == 0:
                    continue
                self.y[t] = self.y_k[nin_count : nin_count + model.ng]
                nin_count += model.ng

        if (
            self.method == "CustomOSQP"
            or self.method == "OSQP"
            or self.method == "HPIPM_DENSE"
            or self.method == "HPIPM_OCP"
        ):
            nin_count = self.n_eq
            self.lag_mul[0] = np.zeros(self.problem.runningModels[0].state.ndx)
            for t in range(self.problem.T + 1):
                if t < self.problem.T:
                    model = self.problem.runningModels[t]
                else:
                    model = self.problem.terminalModel
                if t < self.problem.T:
                    self.lag_mul[t + 1] = -self.y_k[t * self.ndx : (t + 1) * self.ndx]
                if model.ng == 0:
                    continue
                self.y[t] = self.y_k[nin_count : nin_count + model.ng]
                nin_count += model.ng

        self.dx[0] = np.zeros(self.ndx)
        for t in range(self.problem.T):
            self.dx[t + 1] = res[t * self.ndx : (t + 1) * self.ndx]
            index_u = self.problem.T * self.ndx + t * self.nu
            self.du[t] = res[index_u : index_u + self.nu]

        self.x_grad_norm = np.linalg.norm(self.dx) / (self.problem.T + 1)
        self.u_grad_norm = np.linalg.norm(self.du) / self.problem.T

    def acceptStep(self, alpha):
        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            self.xs_try[t] = model.state.integrate(self.xs[t], alpha * self.dx[t])
            self.us_try[t] = self.us[t] + alpha * self.du[t]
        self.xs_try[-1] = model.state.integrate(
            self.xs[-1], alpha * self.dx[-1]
        )  ## terminal state update

        self.setCandidate(self.xs_try, self.us_try, False)

    def solve(
        self, init_xs=None, init_us=None, maxiter=1000, isFeasible=False, regInit=None
    ):
        # ___________________ Initialize ___________________#
        self.max_qp_iters = maxiter
        if init_xs is None or len(init_xs) < 1:
            init_xs = [self.problem.x0.copy() for m in self.models()]
        if init_us is None or len(init_us) < 1:
            init_us = [np.zeros(m.nu) for m in self.problem.runningModels]

        init_xs[0][:] = self.problem.x0.copy()  # Initial condition guess must be x0
        self.setCandidate(init_xs, init_us, False)
        self.computeDirectionFullQP(KKT=False)
        self.acceptStep(alpha=1.0)
        # self.reset_params()

    def allocateDataQP(self):
        #
        self.xs_try = [np.zeros(m.state.nx) for m in self.models()]
        self.xs_try[0][:] = self.problem.x0.copy()
        self.us_try = [np.zeros(m.nu) for m in self.problem.runningModels]
        #
        self.dx = [np.zeros(m.state.ndx) for m in self.models()]
        self.du = [np.zeros(m.nu) for m in self.problem.runningModels]
        #
        self.y = [np.zeros(m.ng) for m in self.problem.runningModels] + [
            np.zeros(self.problem.terminalModel.ng)
        ]
        self.lag_mul = [np.zeros(m.state.ndx) for m in self.models()]
        #
        self.dz_relaxed = [np.zeros(m.ng) for m in self.problem.runningModels] + [
            np.zeros(self.problem.terminalModel.ng)
        ]
        #
        self.x_grad = [np.zeros(m.state.ndx) for m in self.models()]
        self.u_grad = [np.zeros(m.nu) for m in self.problem.runningModels]

        self.gap = [np.zeros(m.state.ndx) for m in self.models()]  # gaps
        self.gap_try = [
            np.zeros(m.state.ndx) for m in self.models()
        ]  # gaps for line search

        self.x_grad_norm = 0
        self.u_grad_norm = 0
        self.gap_norm = 0
        self.cost = 0
        self.cost_try = 0
        self.constraint_norm = 0
        self.constraint_norm_try = 0
        #
        self.nx = self.problem.terminalModel.state.nx
        self.ndx = self.problem.terminalModel.state.ndx
        self.nu = self.problem.runningModels[0].nu

        n_eq = (
            sum([m.nh for m in self.problem.runningModels])
            + self.problem.terminalModel.nh
        )

        assert n_eq == 0



"""__init__

License: BSD 3-Clause License
Copyright (C) 2023, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.
"""

import crocoddyl
import example_robot_data
import matplotlib.pyplot as plt
import mim_solvers
import numpy as np

# # # # # # # # # # # # # # #
###       LOAD ROBOT      ###
# # # # # # # # # # # # # # #

robot = example_robot_data.load("ur5")
model = robot.model
nq = model.nq
nv = model.nv
nu = nv
q0 = np.array([0, 0, 0, 0, 0, 0])
v0 = np.zeros(nv)
x0 = np.concatenate([q0, v0]).copy()

# # # # # # # # # # # # # # #
###  SETUP CROCODDYL OCP  ###
# # # # # # # # # # # # # # #

# State and actuation model
state = crocoddyl.StateMultibody(model)
actuation = crocoddyl.ActuationModelFull(state)

# Create cost terms
# Control regularization cost
uResidual = crocoddyl.ResidualModelControlGrav(state)
uRegCost = crocoddyl.CostModelResidual(state, uResidual)
# State regularization cost
xResidual = crocoddyl.ResidualModelState(state, x0)
xRegCost = crocoddyl.CostModelResidual(state, xResidual)
# endeff frame translation cost
endeff_frame_id = model.getFrameId("tool0")
endeff_translation = np.array([0.4, 0.4, 0.4])
frameTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
    state, endeff_frame_id, endeff_translation
)
frameTranslationCost = crocoddyl.CostModelResidual(state, frameTranslationResidual)

# Create contraint on end-effector
frameTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
    state, endeff_frame_id, np.zeros(3)
)
ee_contraint = crocoddyl.ConstraintModelResidual(
    state,
    frameTranslationResidual,
    np.array([-1.0, -1.0, -1.0]),
    np.array([1.0, 0.4, 0.4]),
)

# Create the running models
runningModels = []
dt = 5e-2
T = 40
for t in range(T + 1):
    runningCostModel = crocoddyl.CostModelSum(state)
    # Add costs
    runningCostModel.addCost("stateReg", xRegCost, 1e-1)
    runningCostModel.addCost("ctrlRegGrav", uRegCost, 1e-4)
    if t != T:
        runningCostModel.addCost("translation", frameTranslationCost, 4)
    else:
        runningCostModel.addCost("translation", frameTranslationCost, 40)
    # Define contraints
    constraints = crocoddyl.ConstraintModelManager(state, nu)
    if t != 0:
        constraints.addConstraint("ee_bound", ee_contraint)
    # Create Differential action model
    running_DAM = crocoddyl.DifferentialActionModelFreeFwdDynamics(
        state, actuation, runningCostModel, constraints
    )
    # Apply Euler integration
    running_model = crocoddyl.IntegratedActionModelEuler(running_DAM, dt)
    runningModels.append(running_model)


# Create the shooting problem
problem = crocoddyl.ShootingProblem(x0, runningModels[:-1], runningModels[-1])


# # # # # # # # # # # # #
###     SOLVE OCP     ###
# # # # # # # # # # # # #

# Define warm start
xs = [x0] * (T + 1)
us = [np.zeros(nu)] * T

# Define solver
solver = mim_solvers.SolverCSQP(problem)
solver.termination_tolerance = 1e-4
# solver.with_callbacks = True
solver.setCallbacks([mim_solvers.CallbackVerbose(), mim_solvers.CallbackLogger()])

# Solve
max_iter = 100
solver.solve(xs, us, max_iter)
log = solver.getCallbacks()[-1]
crocoddyl.plotOCSolution(solver.xs, solver.us)
mim_solvers.plotConvergence(log.convergence_data)


x_traj = np.array(solver.xs)
u_traj = np.array(solver.us)
p_traj = np.zeros((len(solver.xs), 3))

for i in range(T + 1):
    robot.framesForwardKinematics(x_traj[i, :nq])
    p_traj[i] = robot.data.oMf[endeff_frame_id].translation


time_lin = np.linspace(0, dt * (T + 1), T + 1)

fig, axs = plt.subplots(nq)
for i in range(nq):
    axs[i].plot(time_lin, x_traj[:, i])
fig.suptitle("State trajectory")


fig, axs = plt.subplots(nq)
for i in range(nq):
    axs[i].plot(time_lin[:-1], u_traj[:, i])
fig.suptitle("Control trajectory")


fig, axs = plt.subplots(3)
for i in range(3):
    axs[i].plot(time_lin, p_traj[:, i])
    axs[i].plot(time_lin[-1], endeff_translation[i], "o")
fig.suptitle("End effector trajectory")
plt.show()

# viewer
WITHDISPLAY = True
if WITHDISPLAY:
    import time

    display = crocoddyl.MeshcatDisplay(robot)
    display.rate = -1
    display.freq = 1
    while True:
        display.displayFromSolver(solver)
        time.sleep(1.0)



"""
License: BSD 3-Clause License
Copyright (C) 2024, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.

This file checks that the python and c++ csqp implementation match on ur5.
"""

import os
import pathlib

import crocoddyl
import example_robot_data
import mim_solvers
import numpy as np

python_path = pathlib.Path(__file__).absolute().parent.parent.parent / "python"
os.sys.path.insert(1, str(python_path))

from csqp import CSQP  # noqa: E402

# # # # # # # # # # # # # # #
###       LOAD ROBOT      ###
# # # # # # # # # # # # # # #

robot = example_robot_data.load("ur5")
model = robot.model
nq = model.nq
nv = model.nv
nu = nv
q0 = np.array([0, 0, 0, 0, 0, 0])
v0 = np.zeros(nv)
x0 = np.concatenate([q0, v0]).copy()

# # # # # # # # # # # # # # #
###  SETUP CROCODDYL OCP  ###
# # # # # # # # # # # # # # #

# State and actuation model
state = crocoddyl.StateMultibody(model)
actuation = crocoddyl.ActuationModelFull(state)

# Create cost terms
# Control regularization cost
uResidual = crocoddyl.ResidualModelControlGrav(state)
uRegCost = crocoddyl.CostModelResidual(state, uResidual)
# State regularization cost
xResidual = crocoddyl.ResidualModelState(state, x0)
xRegCost = crocoddyl.CostModelResidual(state, xResidual)
# endeff frame translation cost
endeff_frame_id = model.getFrameId("tool0")
endeff_translation = np.array([0.4, 0.4, 0.4])
frameTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
    state, endeff_frame_id, endeff_translation
)
frameTranslationCost = crocoddyl.CostModelResidual(state, frameTranslationResidual)

# Create contraint on end-effector
frameTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
    state, endeff_frame_id, np.zeros(3)
)
ee_contraint = crocoddyl.ConstraintModelResidual(
    state,
    frameTranslationResidual,
    np.array([-1.0, -1.0, -1.0]),
    np.array([1.0, 0.4, 0.4]),
)

# Create the running models
runningModels = []
dt = 5e-2
T = 40
for t in range(T + 1):
    runningCostModel = crocoddyl.CostModelSum(state)
    # Add costs
    runningCostModel.addCost("stateReg", xRegCost, 1e-1)
    runningCostModel.addCost("ctrlRegGrav", uRegCost, 1e-4)
    if t != T:
        runningCostModel.addCost("translation", frameTranslationCost, 4)
    else:
        runningCostModel.addCost("translation", frameTranslationCost, 40)
    # Define contraints
    constraints = crocoddyl.ConstraintModelManager(state, nu)
    if t != 0:
        constraints.addConstraint("ee_bound", ee_contraint)
    # Create Differential action model
    running_DAM = crocoddyl.DifferentialActionModelFreeFwdDynamics(
        state, actuation, runningCostModel, constraints
    )
    # Apply Euler integration
    running_model = crocoddyl.IntegratedActionModelEuler(running_DAM, dt)
    runningModels.append(running_model)


# Create the shooting problem
problem = crocoddyl.ShootingProblem(x0, runningModels[:-1], runningModels[-1])


# # # # # # # # # # # # #
###     SOLVE OCP     ###
# # # # # # # # # # # # #

# Define warm start
xs_init_1 = [x0] * (T + 1)
us_init_1 = [np.zeros(nu)] * T

xs_init_2 = [x0] * (T + 1)
us_init_2 = [np.zeros(nu)] * T

ddp1 = mim_solvers.SolverCSQP(problem)
ddp2 = CSQP(problem, "StagewiseQP")

ddp1.with_callbacks = True
ddp2.with_callbacks = True

ddp1.use_filter_line_search = True
ddp2.use_filter_line_search = True

max_sqp_iter = 4

ddp1.filter_size = max_sqp_iter
ddp2.filter_size = max_sqp_iter

ddp1.extra_iteration_for_last_kkt = True
ddp2.extra_iteration_for_last_kkt = True

termination_tolerance = 1e-8
ddp1.termination_tolerance = termination_tolerance
ddp2.termination_tolerance = termination_tolerance

ddp1.max_qp_iters = 1000
ddp2.max_qp_iters = 1000

ddp1.eps_abs = 1e-4
ddp1.eps_rel = 0.0

ddp2.eps_abs = 1e-4
ddp2.eps_rel = 0.0

# Remove regularization in cpp solver
ddp1.remove_reg = True


set_tol = 1e-8

for reset_rho in [True, False]:
    for reset_y in [True, False]:
        ddp1.reset_rho = reset_rho
        ddp1.reset_y = reset_y

        ddp2.reset_rho = reset_rho
        ddp2.reset_y = reset_y

        converged = ddp1.solve(xs_init_1, us_init_1, max_sqp_iter)
        converged = ddp2.solve(xs_init_2, us_init_2, max_sqp_iter)

        ##### UNIT TEST #####################################
        assert np.linalg.norm(np.array(ddp1.xs) - np.array(ddp2.xs)) < set_tol, (
            "Test failed"
        )
        assert np.linalg.norm(np.array(ddp1.us) - np.array(ddp2.us)) < set_tol, (
            "Test failed"
        )

        assert (
            np.linalg.norm(np.array(ddp1.dx_tilde) - np.array(ddp2.dx_tilde)) < set_tol
        ), "Test failed"
        assert (
            np.linalg.norm(np.array(ddp1.du_tilde) - np.array(ddp2.du_tilde)) < set_tol
        ), "Test failed"

        assert (
            np.linalg.norm(np.array(ddp1.lag_mul) - np.array(ddp2.lag_mul)) < set_tol
        ), "Test failed"

        assert ddp1.qp_iters == ddp2.qp_iters



"""
License: BSD 3-Clause License
Copyright (C) 2024, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.

This file checks that the python dense ADMM implementation matches the stagewise implementations of python and c++.
"""

import os
import pathlib

import mim_solvers
import numpy as np

python_path = pathlib.Path(__file__).absolute().parent.parent.parent / "python"
os.sys.path.insert(1, str(python_path))

from csqp import CSQP  # noqa: E402
from problems import create_clqr_problem  # noqa: E402

LINE_WIDTH = 100

print(" TEST Stagewise ADMM PROBLEM ".center(LINE_WIDTH, "-"))

problem, xs_init, us_init = create_clqr_problem()

ddp1 = mim_solvers.SolverCSQP(problem)
ddp1.remove_reg = True
ddp1.reset_rho = True  # To-do: make sure that the python has the same default arguments
ddp1.reset_y = True


ddp2 = CSQP(problem, "StagewiseQP")
ddp3 = CSQP(problem, "StagewiseQPKKT")

ddp1.with_callbacks = True
ddp2.with_callbacks = True
ddp3.with_callbacks = True

termination_tolerance = 1e-4
ddp1.termination_tolerance = termination_tolerance
ddp2.termination_tolerance = termination_tolerance
ddp3.termination_tolerance = termination_tolerance

max_qp_iters = 1000
ddp1.max_qp_iters = max_qp_iters
ddp2.max_qp_iters = max_qp_iters
ddp3.max_qp_iters = max_qp_iters

eps_abs = 1e-8
eps_rel = 0.0
ddp1.eps_abs = eps_abs
ddp2.eps_abs = eps_abs
ddp3.eps_abs = eps_abs
ddp1.eps_rel = eps_rel
ddp2.eps_rel = eps_rel
ddp3.eps_rel = eps_rel


converged = ddp1.solve(xs_init, us_init, 1)
converged = ddp2.solve(xs_init, us_init, 1)
converged = ddp3.solve(xs_init, us_init, 1)


assert ddp1.qp_iters == ddp2.qp_iters
set_tol = 1e-6
assert np.linalg.norm(np.array(ddp1.dx_tilde) - np.array(ddp2.dx_tilde)) < set_tol, (
    "Test failed"
)
assert np.linalg.norm(np.array(ddp1.du_tilde) - np.array(ddp2.du_tilde)) < set_tol, (
    "Test failed"
)


assert np.linalg.norm(ddp1.norm_primal - np.array(ddp2.norm_primal)) < set_tol, (
    "Test failed"
)
assert np.linalg.norm(ddp1.norm_dual - np.array(ddp2.norm_dual)) < set_tol, (
    "Test failed"
)
assert (
    np.linalg.norm(ddp1.norm_primal_rel - np.array(ddp2.norm_primal_rel)) < set_tol
), "Test failed"
assert np.linalg.norm(ddp1.norm_dual_rel - np.array(ddp2.norm_dual_rel)) < set_tol, (
    "Test failed"
)

assert np.linalg.norm(ddp1.rho_sparse - np.array(ddp2.rho_sparse)) < set_tol, (
    "Test failed"
)


for t in range(len(ddp1.rho_vec)):
    assert np.linalg.norm(ddp1.rho_vec[t] - ddp2.rho_vec[t]) < set_tol, "Test failed"
    assert np.linalg.norm(ddp1.y[t] - ddp2.y[t]) < set_tol, "Test failed"
    assert np.linalg.norm(ddp1.z[t] - ddp2.z[t]) < set_tol, "Test failed"



"""__init__
License: BSD 3-Clause License
Copyright (C) 2023, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.
"""

import collections

import crocoddyl
import numpy as np

LINE_WIDTH = 100


class SQP_CPP(crocoddyl.SolverFDDP):
    def __init__(
        self, shootingProblem, use_filter_line_search=True, with_callbacks=False
    ):
        crocoddyl.SolverFDDP.__init__(self, shootingProblem)
        self.mu = 1.0
        self.termination_tolerance = 1e-6
        self.with_callbacks = with_callbacks
        self.use_filter_line_search = use_filter_line_search
        self.filter_size = 1

        self.allocateData()

    def calc(self):
        self.calcDiff()
        self.gap_norm = sum(np.linalg.norm(self.fs, 1, axis=1))
        self.merit = self.cost + self.mu * self.gap_norm
        # print(self.gap_norm)

    def computeDirection(self, kkt_check=True):
        # print("using Python")
        self.calc()
        if kkt_check:
            self.KKT_check()
            if self.KKT < self.termination_tolerance:
                return True

        self.backwardPass()
        self.computeUpdates()
        return False

    def computeUpdates(self):
        """computes step updates dx and du"""
        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            # here we compute the direction
            self.lag_mul[t] = self.Vxx[t] @ (self.dx[t] - self.fs[t]) + self.Vx[t]
            self.du[t][:] = -self.K[t].dot(self.dx[t]) - self.k[t]
            A = data.Fx
            B = data.Fu
            if len(data.Fu.shape) == 1:
                bl = -B.dot(self.k[t][0])
                BL = -B.reshape(B.shape[0], 1) @ self.K[t].reshape(1, B.shape[0])
            else:
                bl = -B @ self.k[t]
                BL = -B @ self.K[t]
            self.dx[t + 1] = (A + BL) @ self.dx[t] + bl + self.fs[t + 1]

        self.lag_mul[-1] = self.Vxx[-1] @ (self.dx[-1] - self.fs[-1]) + self.Vx[-1]
        self.x_grad_norm = sum(np.linalg.norm(self.dx, 1, axis=1)) / self.problem.T
        self.u_grad_norm = sum(np.linalg.norm(self.du, 1, axis=1)) / self.problem.T
        # print("x_norm", self.x_grad_norm, "u_norm", self.u_grad_norm )

    def KKT_check(self):
        self.KKT = 0
        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            self.KKT = max(
                self.KKT,
                max(abs(data.Lx + data.Fx.T @ self.lag_mul[t + 1] - self.lag_mul[t])),
            )
            self.KKT = max(
                self.KKT, max(abs(data.Lu + data.Fu.T @ self.lag_mul[t + 1]))
            )

        self.KKT = max(
            self.KKT, max(abs(self.problem.terminalData.Lx - self.lag_mul[-1]))
        )
        self.KKT = max(self.KKT, max(abs(np.array(self.fs).flatten())))

    def tryStep(self, alpha):
        # print("using python")
        self.merit_try = 0
        self.cost_try = 0

        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            self.xs_try[t] = model.state.integrate(self.xs[t], alpha * self.dx[t])
            self.us_try[t] = self.us[t] + alpha * self.du[t]
        self.xs_try[-1] = model.state.integrate(
            self.xs[-1], alpha * self.dx[-1]
        )  ## terminal state update

        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            model.calc(data, self.xs_try[t], self.us_try[t])
            self.gap_try[t] = model.state.diff(self.xs_try[t + 1], data.xnext)  # gaps
            self.cost_try += data.cost

        self.problem.terminalModel.calc(self.problem.terminalData, self.xs_try[-1])
        self.cost_try += self.problem.terminalData.cost

        self.gap_norm_try = (
            sum(np.linalg.norm(self.gap_try, 1, axis=1)) / self.problem.T
        )

        self.merit_try = self.cost_try + self.mu * self.gap_norm_try
        # print("cost_try", self.cost_try, "gaps_try", self.gap_norm_try, "merit try", self.merit_try)

    def acceptStep(self):
        self.setCandidate(self.xs_try, self.us_try, False)

    def solve(
        self, init_xs=None, init_us=None, maxiter=100, isFeasible=False, regInit=None
    ):
        if init_xs is None or len(init_xs) < 1:
            init_xs = [self.problem.x0.copy() for m in self.models()]
        if init_us is None or len(init_us) < 1:
            init_us = [np.zeros(m.nu) for m in self.problem.runningModels]

        init_xs[0] = self.problem.x0.copy()  # Initial condition guess must be x0

        self.setCandidate(init_xs, init_us, False)

        alpha = None
        self.cost_list = collections.deque(
            [np.inf] * self.filter_size, maxlen=self.filter_size
        )
        self.gap_list = collections.deque(
            [np.inf] * self.filter_size, maxlen=self.filter_size
        )

        if self.with_callbacks:
            headings = ["iter", "merit", "cost", "grad", "step", "||gaps||", "KKT"]
            print("{:>3} {:>9} {:>10} {:>11} {:>8} {:>11} {:>8}".format(*headings))

        for iter in range(maxiter):
            self.computeDirection()
            self.gap_list.append(self.gap_norm)
            self.cost_list.append(self.cost)
            alpha = 1.0
            max_search = 10
            for k in range(max_search):
                self.tryStep(alpha)
                if self.use_filter_line_search:
                    is_worse_than_memory = False
                    count = 0
                    while (
                        count < self.filter_size
                        and not is_worse_than_memory
                        and count <= iter
                    ):
                        is_worse_than_memory = (
                            self.cost_list[self.filter_size - 1 - count] < self.cost_try
                            and self.gap_list[self.filter_size - 1 - count]
                            < self.gap_norm_try
                        )
                        count += 1

                    if not is_worse_than_memory:
                        self.acceptStep()
                        break
                else:
                    if k == max_search - 1:
                        print("No improvement")
                        return False
                    if self.merit > self.merit_try:
                        self.acceptStep()
                        break

                alpha *= 0.5

            if self.KKT < self.termination_tolerance:
                return True

            if self.with_callbacks:
                print(
                    "{:>4} {:.5e} {:.5e} {:.5e} {:.4f} {:.5e} {:.5e}".format(
                        iter,
                        float(self.merit),
                        self.cost,
                        self.x_grad_norm + self.u_grad_norm,
                        alpha,
                        self.gap_norm,
                        self.KKT,
                    )
                )

        return True

    def models(self):
        mod = [m for m in self.problem.runningModels]
        mod += [self.problem.terminalModel]
        return mod

    def allocateData(self):
        self.xs_try = [np.zeros(m.state.nx) for m in self.models()]
        self.xs_try[0][:] = self.problem.x0.copy()
        self.us_try = [np.zeros(m.nu) for m in self.problem.runningModels]
        #
        self.dx = [np.zeros(m.state.ndx) for m in self.models()]
        self.du = [np.zeros(m.nu) for m in self.problem.runningModels]

        self.lag_mul = [np.zeros(m.state.ndx) for m in self.models()]

        self.gap = [np.zeros(m.state.ndx) for m in self.models()]  # gaps
        self.gap_try = [
            np.zeros(m.state.ndx) for m in self.models()
        ]  # gaps for line search

        self.merit = 0
        self.merit_old = 0
        self.x_grad_norm = 0
        self.u_grad_norm = 0
        self.gap_norm = 0
        self.cost = 0
        self.cost_try = 0
        self.expected_decrease = 0



## Implementation of the sequential constrained lqr
## Author : Avadesh Meduri and Armand Jordana
## Date : 9/3/2022

import collections

import numpy as np
import scipy.linalg as scl
from qp_solvers.qpsolvers import QPSolvers
from qp_solvers.stagewise_qp import StagewiseADMM


def pp(s):
    return np.format_float_scientific(s, exp_digits=2, precision=4)


def rev_enumerate(l):  # noqa: E741
    return reversed(list(enumerate(l)))


def raiseIfNan(A, error=None):
    if error is None:
        error = scl.LinAlgError("NaN in array")
    if np.any(np.isnan(A)) or np.any(np.isinf(A)) or np.any(abs(np.asarray(A)) > 1e30):
        raise error


class CSQP(StagewiseADMM, QPSolvers):
    def __init__(
        self,
        shootingProblem,
        method,
        use_filter_line_search=True,
        with_callbacks=False,
        qp_with_callbacks=False,
    ):
        if method == "StagewiseQP":
            StagewiseADMM.__init__(self, shootingProblem, verboseQP=qp_with_callbacks)
            self.using_qp = 0
        else:
            QPSolvers.__init__(
                self, shootingProblem, method, verboseQP=qp_with_callbacks
            )
            self.using_qp = 1

        self.mu1 = 1e1
        self.mu2 = 1e1
        self.lag_mul_inf_norm_coef = 1.01
        self.lag_mul_inf_norm = 0
        self.termination_tolerance = 1e-6

        self.iter = 0
        self.use_filter_line_search = use_filter_line_search
        self.filter_size = 1
        self.with_callbacks = with_callbacks
        self.extra_iteration_for_last_kkt = False

    def models(self):
        mod = [m for m in self.problem.runningModels]
        mod += [self.problem.terminalModel]
        return mod

    def tryStep(self, alpha):
        """
        This function tries the step
        """

        self.merit_try = 0
        self.constraint_norm_try = 0
        self.cost_try = 0
        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            self.xs_try[t] = model.state.integrate(self.xs[t], alpha * self.dx[t])
            self.us_try[t] = self.us[t] + alpha * self.du[t]
        self.xs_try[-1] = model.state.integrate(
            self.xs[-1], alpha * self.dx[-1]
        )  ## terminal state update

        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            model.calc(data, self.xs_try[t], self.us_try[t])

            self.gap_try[t] = model.state.diff(self.xs_try[t + 1], data.xnext)  # gaps
            self.cost_try += data.cost

            self.constraint_norm_try += np.linalg.norm(
                np.clip(model.g_lb - data.g, 0, np.inf), 1
            )
            self.constraint_norm_try += np.linalg.norm(
                np.clip(data.g - model.g_ub, 0, np.inf), 1
            )

        self.problem.terminalModel.calc(self.problem.terminalData, self.xs_try[-1])
        self.constraint_norm_try += np.linalg.norm(
            np.clip(
                self.problem.terminalModel.g_lb - self.problem.terminalData.g, 0, np.inf
            ),
            1,
        )
        self.constraint_norm_try += np.linalg.norm(
            np.clip(
                self.problem.terminalData.g - self.problem.terminalModel.g_ub, 0, np.inf
            ),
            1,
        )

        self.cost_try += self.problem.terminalData.cost

        self.gap_norm_try = sum(np.linalg.norm(self.gap_try, 1, axis=1))

        if self.mu1 < 0 or self.mu2 < 0:
            self.merit_try = (
                self.cost_try
                + self.lag_mul_inf_norm_coef
                * self.lag_mul_inf_norm
                * (self.gap_norm_try + self.constraint_norm_try)
            )
        else:
            self.merit_try = (
                self.cost_try
                + self.mu1 * self.gap_norm_try
                + self.mu2 * self.constraint_norm_try
            )

    def LQ_problem_KKT_check(self):
        KKT = 0
        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            Cx, Cu = data.Gx, data.Gu
            if t == 0:
                lu = (
                    data.Luu @ self.du[t]
                    + data.Lxu.T @ self.dx[t]
                    + data.Lu
                    + data.Fu.T @ self.lag_mul[t + 1]
                    + Cu.T @ self.y[t]
                )
                KKT = max(KKT, max(abs(lu)))
                continue

            lx = (
                data.Lxx @ self.dx[t]
                + data.Lxu @ self.du[t]
                + data.Lx
                + data.Fx.T @ self.lag_mul[t + 1]
                - self.lag_mul[t]
                + Cx.T @ self.y[t]
            )
            lu = (
                data.Luu @ self.du[t]
                + data.Lxu.T @ self.dx[t]
                + data.Lu
                + data.Fu.T @ self.lag_mul[t + 1]
                + Cu.T @ self.y[t]
            )
            KKT = max(KKT, max(abs(lx)), max(abs(lu)))

            l1 = np.max(
                np.abs(
                    np.clip(
                        model.g_lb - Cx @ self.dx[t] - Cu @ self.du[t] - data.g,
                        0,
                        np.inf,
                    )
                )
            )
            l2 = np.max(
                np.abs(
                    np.clip(
                        Cx @ self.dx[t] + Cu @ self.du[t] + data.g - model.g_ub,
                        0,
                        np.inf,
                    )
                )
            )
            l3 = np.max(
                np.abs(
                    self.dx[t + 1]
                    - data.Fx @ self.dx[t]
                    - data.Fu @ self.du[t]
                    - self.gap[t]
                )
            )
            KKT = max(KKT, l1, l2, l3)

        model = self.problem.terminalModel
        data = self.problem.terminalData
        Cx = data.Gx
        if model.ng != 0:
            l1 = np.max(
                np.abs(np.clip(model.g_lb - Cx @ self.dx[-1] - data.g, 0, np.inf))
            )
            l2 = np.max(
                np.abs(np.clip(Cx @ self.dx[-1] + data.g - model.g_ub, 0, np.inf))
            )
        lx = (
            self.problem.terminalData.Lxx @ self.dx[-1]
            + self.problem.terminalData.Lx
            - self.lag_mul[-1]
            + Cx.T @ self.y[-1]
        )
        KKT = max(KKT, l1, l2)
        KKT = max(KKT, max(abs(lx)))

        # Note that for this test to pass, the tolerance of the QP should be low.
        # assert KKT < 1e-6
        print("\n This should match the tolerance of the QP solver ", KKT)

    def KKT_check(self):
        if not self.using_qp:
            for t, (model, data) in enumerate(
                zip(self.problem.runningModels, self.problem.runningDatas)
            ):
                self.lag_mul[t] = self.Vxx[t] @ self.dx_tilde[t] + self.Vx[t]
            self.lag_mul[-1] = self.Vxx[-1] @ self.dx_tilde[-1] + self.Vx[-1]
        self.KKT = 0
        for t, data in enumerate(self.problem.runningDatas):
            Cx, Cu = data.Gx, data.Gu
            if t == 0:
                lu = data.Lu + data.Fu.T @ self.lag_mul[t + 1] + Cu.T @ self.y[t]
                self.KKT = max(self.KKT, max(abs(lu)))
                continue
            Cx, Cu = data.Gx, data.Gu
            lx = (
                data.Lx
                + data.Fx.T @ self.lag_mul[t + 1]
                - self.lag_mul[t]
                + Cx.T @ self.y[t]
            )
            lu = data.Lu + data.Fu.T @ self.lag_mul[t + 1] + Cu.T @ self.y[t]
            self.KKT = max(self.KKT, max(abs(lx)), max(abs(lu)))

        Cx = self.problem.terminalData.Gx
        lx = self.problem.terminalData.Lx - self.lag_mul[-1] + Cx.T @ self.y[-1]
        self.KKT = max(self.KKT, max(abs(lx)))
        self.KKT = max(self.KKT, max(abs(np.array(self.gap).flatten())))
        self.KKT = max(self.KKT, self.constraint_norm)

    def solve(
        self, init_xs=None, init_us=None, maxiter=100, isFeasible=False, regInit=None
    ):
        # ___________________ Initialize ___________________#
        if init_xs is None or len(init_xs) < 1:
            init_xs = [self.problem.x0.copy() for m in self.models()]
        if init_us is None or len(init_us) < 1:
            init_us = [np.zeros(m.nu) for m in self.problem.runningModels]

        init_xs[0] = self.problem.x0.copy()  # Initial condition guess must be x0

        self.gap_list = collections.deque(
            self.filter_size * [np.inf], maxlen=self.filter_size
        )
        self.cost_list = collections.deque(
            self.filter_size * [np.inf], maxlen=self.filter_size
        )
        self.constraint_list = collections.deque(
            self.filter_size * [np.inf], maxlen=self.filter_size
        )

        self.setCandidate(init_xs, init_us, False)

        alpha = None
        if self.with_callbacks:
            headings = [
                "iter",
                "merit",
                "cost",
                "||gaps||",
                "||Constraint||",
                "||(dx,du)||",
                "step",
                "KKT",
                "QP Iters",
            ]

            print(
                "{:>3} {:>7} {:>8} {:>8} {:>8} {:>11} {:>11} {:>8} {:>8}".format(
                    *headings
                )
            )
        for iter in range(maxiter):
            # self.iter = iter
            self.calc(True)
            if self.using_qp:
                self.computeDirectionFullQP()
            else:
                if iter == 0 and not self.reset_rho:
                    self.reset_rho_vec()

                self.computeDirection()
            # self.LQ_problem_KKT_check()

            self.KKT_check()
            if self.KKT < self.termination_tolerance:
                if self.with_callbacks:
                    print(
                        "{:>4} {:.4e} {:.4e} {:.4e} {:>4} {:.4e} {:>4} {:.4e} {:>4}".format(
                            "END",
                            float(self.merit),
                            self.cost,
                            self.gap_norm,
                            self.constraint_norm,
                            self.x_grad_norm + self.u_grad_norm,
                            " ---- ",
                            self.KKT,
                            self.qp_iters,
                        )
                    )
                return True

            self.gap_list.append(self.gap_norm)
            self.cost_list.append(self.cost)
            self.constraint_list.append(self.constraint_norm)
            alpha = 1.0
            max_search = 10
            for k in range(max_search):
                self.tryStep(alpha)
                if self.use_filter_line_search:
                    is_worse_than_memory = False
                    count = 0
                    while (
                        count < self.filter_size
                        and not is_worse_than_memory
                        and count <= iter
                    ):
                        is_worse_than_memory = (
                            self.cost_list[self.filter_size - 1 - count] < self.cost_try
                            and self.gap_list[self.filter_size - 1 - count]
                            < self.gap_norm_try
                        )
                        count += 1

                    if not is_worse_than_memory:
                        self.acceptStep(alpha)
                        break
                else:
                    if k == max_search - 1:
                        print("No improvement")
                        return False
                    if self.merit > self.merit_try:
                        self.acceptStep(alpha)
                        break

                alpha *= 0.5

            if self.with_callbacks:
                print(
                    "{:>4} {:.4e} {:.4e} {:.4e} {:.4e} {:.4e} {:.4f} {:.4e} {:>4}".format(
                        iter + 1,
                        float(self.merit),
                        self.cost,
                        self.gap_norm,
                        self.constraint_norm,
                        self.x_grad_norm + self.u_grad_norm,
                        alpha,
                        self.KKT,
                        self.qp_iters,
                    )
                )

        if self.extra_iteration_for_last_kkt:
            self.calc(True)
            if self.using_qp:
                self.computeDirectionFullQP()
            else:
                self.computeDirection()
            # if(self.BENCHMARK):
            #     self.check_qp_convergence()
            self.KKT_check()
            if self.with_callbacks:
                print(
                    "{:>4} {:.4e} {:.4e} {:.4e} {:.4e} {:.4e} {:>4} {:.4e} {:>4}".format(
                        "END",
                        float(self.merit),
                        self.cost,
                        self.gap_norm,
                        self.constraint_norm,
                        self.x_grad_norm + self.u_grad_norm,
                        " ---- ",
                        self.KKT,
                        " ---- ",
                    )
                )
            if self.KKT < self.termination_tolerance:
                return True

        return False



"""
License: BSD 3-Clause License
Copyright (C) 2024, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.

This file checks that the python and c++ sqp implementation match without regularization.
"""

import os
import pathlib

import mim_solvers
import numpy as np

python_path = pathlib.Path(__file__).absolute().parent.parent.parent / "python"
os.sys.path.insert(1, str(python_path))

from problems import (  # noqa: E402
    create_double_pendulum_problem,
    create_quadrotor_problem,
    create_taichi,
    create_unconstrained_ur5,
)
from sqp import SQP  # noqa: E402

# Solver params
MAXITER = 100
TOL = 1e-4
with_callbacks = True
FILTER_SIZE = MAXITER

# Create 1 solver of each type for each problem

problems = [
    create_double_pendulum_problem(),
    create_quadrotor_problem(),
    create_unconstrained_ur5(),
    create_taichi(),
]

for problem, xs_init, us_init in problems:
    x0 = problem.x0.copy()
    # Create solver SQP (MS)
    solverSQP = mim_solvers.SolverSQP(problem)
    solverSQP.termination_tolerance = TOL
    solverSQP.use_filter_line_search = True
    solverSQP.filter_size = MAXITER
    solverSQP.with_callbacks = with_callbacks
    solverSQP.reg_min = 0.0  # This turns of regularization completely.
    reginit = 0.0

    # Create python solver
    pysolverSQP = SQP(problem, with_callbacks=with_callbacks)
    pysolverSQP.termination_tolerance = TOL
    pysolverSQP.use_filter_line_search = True
    pysolverSQP.filter_size = MAXITER

    # SQP
    solverSQP.solve(xs_init, us_init, MAXITER, False, reginit)

    pysolverSQP.solve(xs_init, us_init, MAXITER)

##### UNIT TEST #####################################

set_tol = 1e-4

for t in range(problem.T - 1, 0, -1):
    assert (
        np.linalg.norm(pysolverSQP.L[t] + solverSQP.K[t], 1)
        < (np.size(pysolverSQP.L[t])) * set_tol
    )
    assert (
        np.linalg.norm(pysolverSQP.l[t] + solverSQP.k[t], 1) / (len(pysolverSQP.l[t]))
        < set_tol
    )

assert (
    np.linalg.norm(np.array(pysolverSQP.xs) - np.array(solverSQP.xs)) / (problem.T + 1)
    < set_tol
), "Test failed"
assert (
    np.linalg.norm(np.array(pysolverSQP.us) - np.array(solverSQP.us)) / problem.T
    < set_tol
), "Test failed"

assert pysolverSQP.KKT - solverSQP.KKT < set_tol, "Test failed"

print("ALL UNIT TEST PASSED .....")



"""
License: BSD 3-Clause License
Copyright (C) 2024, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.

This file checks that all methods converge in one iteration on a LQR problem.
"""

import os
import pathlib

import mim_solvers
import numpy as np

python_path = pathlib.Path(__file__).absolute().parent.parent.parent / "python"
os.sys.path.insert(1, str(python_path))

from problems import create_lqr_problem  # noqa: E402
from sqp import SQP  # noqa: E402
from sqp_cpp import SQP_CPP  # noqa: E402

LINE_WIDTH = 100


print("TEST 1: python SQP = mim_solvers SQP".center(LINE_WIDTH, "-"))

problem, xs_init, us_init = create_lqr_problem()
ddp0 = SQP(problem)
ddp1 = SQP_CPP(problem)
ddp2 = mim_solvers.SolverSQP(problem)

ddp0.with_callbacks = True
ddp1.with_callbacks = True
ddp2.with_callbacks = True

ddp0.termination_tolerance = 1e-6
ddp1.termination_tolerance = 1e-6
ddp2.termination_tolerance = 1e-6


converged = ddp0.solve(xs_init, us_init, 10)
converged = ddp1.solve(xs_init, us_init, 10)
converged = ddp2.solve(xs_init, us_init, 10)

tol = 1e-4
assert np.linalg.norm(np.array(ddp0.xs) - np.array(ddp2.xs)) < tol, "Test failed"
assert np.linalg.norm(np.array(ddp0.us) - np.array(ddp2.us)) < tol, "Test failed"
assert np.linalg.norm(np.array(ddp1.xs) - np.array(ddp2.xs)) < tol, "Test failed"
assert np.linalg.norm(np.array(ddp1.us) - np.array(ddp2.us)) < tol, "Test failed"


assert ddp0.iter == 0
assert ddp1.iter == 0
assert ddp2.iter == 1


print("TEST PASSED".center(LINE_WIDTH, "-"))
print("\n")



## Custom osqp implementation
## Author : Avadesh Meduri
## Date : 31/03/2023

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import spsolve

np.set_printoptions(precision=2)
# np.format_float_scientific(self.r_prim, exp_digits=2, precision =2)


def pp(s):
    return np.format_float_scientific(s, exp_digits=2, precision=4)


class StagewiseQPKKT:
    def __init__(self):
        self.sigma_boyd = 1e-6
        self.rho_boyd = 1e-1
        self.rho_min = 1e-6
        self.rho_max = 1e3
        self.alpha_boyd = 1.6

        self.adaptive_rho_tolerance = 5
        self.rho_update_interval = 25

    def solve_linear_system_boyd(self):
        rho_mat = self.rho_vec_boyd * np.eye(len(self.rho_vec_boyd))
        A_block_leftcol = sparse.vstack(
            [
                self.P
                + self.A_in.T @ rho_mat @ self.A_in
                + self.sigma_boyd * np.eye(self.n_vars),
                self.A_eq,
            ]
        )
        A_block_rightcol = sparse.vstack(
            [self.A_eq.T, np.zeros((self.n_eq, self.n_eq))]
        )
        A_block = sparse.hstack([A_block_leftcol, A_block_rightcol])
        A_block = sparse.csr_matrix(A_block)

        tmp = self.A_in.T @ (
            np.multiply(
                self.rho_vec_boyd, self.z_k - np.divide(self.y_k, self.rho_vec_boyd)
            )
        )
        b_block = np.hstack((tmp - self.q + self.sigma_boyd * self.x_k, self.b))

        xv_k_1 = spsolve(A_block, b_block)
        self.xtilde_k_1, self.v_k_1 = xv_k_1[: self.n_vars], xv_k_1[self.n_vars :]

    def update_lagrangian_params_boyd(self):
        self.x_k_1 = (
            self.alpha_boyd * self.A_in @ self.xtilde_k_1
            + (1 - self.alpha_boyd) * self.z_k
        )  # relaxed update
        self.z_k_1 = np.clip(
            self.x_k_1 + np.divide(self.y_k, self.rho_vec_boyd), self.lboyd, self.uboyd
        )
        self.y_k_1 = self.y_k + np.multiply(
            self.rho_vec_boyd, (self.x_k_1 - self.z_k_1)
        )

        dual_vec = self.A_in.T @ np.multiply(self.rho_vec_boyd, (self.z_k_1 - self.z_k))
        self.r_dual = max(abs(dual_vec))
        self.x_k, self.z_k, self.y_k = (
            self.xtilde_k_1.copy(),
            self.z_k_1.copy(),
            self.y_k_1.copy(),
        )

        if self.A_in.shape[0] == 0:
            tmp = 0
            self.eps_rel_prim = 1e-10
            self.eps_rel_dual = 1e-10
        else:
            tmp = max(abs(self.A_in @ self.x_k - self.z_k))
            self.eps_rel_prim = max(abs(np.hstack((self.A_in @ self.x_k, self.z_k))))
            self.eps_rel_dual = max(abs(self.A_in.T @ self.y_k))

        self.r_prim = max(tmp, max(abs(self.A_eq @ self.x_k - self.b)))

        ## This is the OSQP dual computation
        # self.r_dual = max(abs(self.P @ self.x_k + self.q + self.A_in.T @ self.y_k + self.A_eq.T @ self.v_k_1))

        # tmp = max(abs(self.q))
        # tmp2 = max(abs(self.A_in.T @ self.y_k))
        # tmp3 = max(abs(self.P @ self.x_k))
        # tmp4 = max(abs(self.A_eq.T @ self.v_k_1))
        # self.eps_rel_dual = max(tmp, tmp2)
        # self.eps_rel_dual = max(self.eps_rel_dual, tmp3)
        # self.eps_rel_dual = max(self.eps_rel_dual, tmp4)

    def update_rho_boyd(self, iter):
        scale = (self.r_prim * self.eps_rel_dual) / (self.r_dual * self.eps_rel_prim)
        scale = np.sqrt(scale)
        self.scale_boyd = scale
        self.rho_estimate_boyd = scale * self.rho_boyd
        self.rho_estimate_boyd = min(
            max(self.rho_estimate_boyd, self.rho_min), self.rho_max
        )

        if (iter) % self.rho_update_interval == 0 and iter > 1:
            if (
                self.rho_estimate_boyd > self.rho_boyd * self.adaptive_rho_tolerance
                or self.rho_estimate_boyd < self.rho_boyd / self.adaptive_rho_tolerance
            ):
                self.rho_boyd = self.rho_estimate_boyd
                for i in range(len(self.lboyd)):
                    if self.lboyd[i] == -np.inf and self.uboyd[i] == np.inf:
                        self.rho_vec_boyd[i] = self.rho_min
                    elif abs(self.lboyd[i] - self.uboyd[i]) < 1e-4:
                        self.rho_vec_boyd[i] = 1e3 * self.rho_boyd
                    elif self.lboyd[i] != self.uboyd[i]:
                        self.rho_vec_boyd[i] = self.rho_boyd
                    else:
                        assert False  # safety check

    def set_rho_boyd(self):
        self.rho_vec_boyd = np.zeros(self.n_in)
        self.rho_boyd = min(max(self.rho_boyd, self.rho_min), self.rho_max)
        for i in range(len(self.lboyd)):
            if self.lboyd[i] == -np.inf and self.uboyd[i] == np.inf:
                self.rho_vec_boyd[i] = self.rho_min
            elif self.lboyd[i] == self.uboyd[i]:
                self.rho_vec_boyd[i] = 1e3 * self.rho_boyd
            elif self.lboyd[i] != self.uboyd[i]:
                self.rho_vec_boyd[i] = self.rho_boyd
        self.rho_estimate_boyd = 0

    def optimize_boyd(self, maxiters=1000):
        self.set_rho_boyd()
        converged = False
        for iter in range(1, maxiters + 1):
            self.solve_linear_system_boyd()
            self.update_lagrangian_params_boyd()
            self.update_rho_boyd(iter)

            eps_prim = self.eps_abs + self.eps_rel * self.eps_rel_prim
            eps_dual = self.eps_abs + self.eps_rel * self.eps_rel_dual

            if (iter) % self.rho_update_interval == 0 and iter > 1:
                if self.r_prim <= eps_prim and self.r_dual <= eps_dual:
                    if self.verboseQP:
                        print(
                            "Iters",
                            iter,
                            "res-primal",
                            pp(self.r_prim),
                            "res-dual",
                            pp(self.r_dual),
                            "optimal rho estimate",
                            pp(self.rho_estimate_boyd),
                            "rho",
                            pp(self.rho_boyd),
                        )
                        print("terminated ... \n")
                    converged = True
                    break
                if self.verboseQP:
                    print(
                        "Iters",
                        iter,
                        "res-primal",
                        pp(self.r_prim),
                        "res-dual",
                        pp(self.r_dual),
                        "optimal rho estimate",
                        pp(self.rho_estimate_boyd),
                        "rho",
                        pp(self.rho_boyd),
                    )
        self.qp_iters = iter
        if not converged and self.verboseQP:
            print("Not Converged ... \n")
            print(
                "Iters",
                iter,
                "res-primal",
                pp(self.r_prim),
                "res-dual",
                pp(self.r_dual),
                "optimal rho estimate",
                pp(self.rho_estimate_boyd),
                "rho",
                pp(self.rho_boyd),
            )
        return self.x_k.copy()



"""
License: BSD 3-Clause License
Copyright (C) 2023, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.
"""

import sys

import crocoddyl
import example_robot_data
import matplotlib.pyplot as plt
import mim_solvers
import numpy as np
import pinocchio
import pinocchio as pin
import utils_solo12

pinRef = pin.LOCAL_WORLD_ALIGNED
FRICTION_CSTR = True
MU = 0.8  # friction coefficient

PLOT_OCP_SOL = False
PLAY_OCP_SOL = True

robot_name = "solo12"
ee_frame_names = ["FL_FOOT", "FR_FOOT", "HL_FOOT", "HR_FOOT"]
solo12 = example_robot_data.ROBOTS[robot_name]()
rmodel = solo12.robot.model
rmodel.type = "QUADRUPED"
rmodel.foot_type = "POINT_FOOT"
rdata = rmodel.createData()

# set contact frame_names and_indices
lfFootId = rmodel.getFrameId(ee_frame_names[0])
rfFootId = rmodel.getFrameId(ee_frame_names[1])
lhFootId = rmodel.getFrameId(ee_frame_names[2])
rhFootId = rmodel.getFrameId(ee_frame_names[3])


q0 = np.array(
    [0.0, 0.0, 0.25, 0.0, 0.0, 0.0, 1.0] + 2 * [0.0, 0.8, -1.6] + 2 * [0.0, -0.8, 1.6]
)

x0 = np.concatenate([q0, np.zeros(rmodel.nv)])

pinocchio.forwardKinematics(rmodel, rdata, q0)
pinocchio.updateFramePlacements(rmodel, rdata)
rfFootPos0 = rdata.oMf[rfFootId].translation
rhFootPos0 = rdata.oMf[rhFootId].translation
lfFootPos0 = rdata.oMf[lfFootId].translation
lhFootPos0 = rdata.oMf[lhFootId].translation

comRef = (rfFootPos0 + rhFootPos0 + lfFootPos0 + lhFootPos0) / 4
comRef[2] = pinocchio.centerOfMass(rmodel, rdata, q0)[2].item()

supportFeetIds = [lfFootId, rfFootId, lhFootId, rhFootId]
supportFeePos = [lfFootPos0, rfFootPos0, lhFootPos0, rhFootPos0]


state = crocoddyl.StateMultibody(rmodel)
actuation = crocoddyl.ActuationModelFloatingBase(state)
nu = actuation.nu


comDes = []

N_ocp = 250
dt = 0.02
T = N_ocp * dt
radius = 0.065
for t in range(N_ocp + 1):
    comDes_t = comRef.copy()
    w = (2 * np.pi) * 0.2
    comDes_t[0] += radius * np.sin(w * t * dt)
    comDes_t[1] += radius * (np.cos(w * t * dt) - 1)
    comDes += [comDes_t]

running_models = []
constraintModels = []
for t in range(N_ocp + 1):
    contactModel = crocoddyl.ContactModelMultiple(state, nu)
    costModel = crocoddyl.CostModelSum(state, nu)

    # Add contact
    for frame_idx in supportFeetIds:
        support_contact = crocoddyl.ContactModel3D(
            state,
            frame_idx,
            np.array([0.0, 0.0, 0.0]),
            pinRef,
            nu,
            np.array([0.0, 0.0]),
        )
        contactModel.addContact(
            rmodel.frames[frame_idx].name + "_contact", support_contact
        )

    # Add state/control reg costs

    state_reg_weight, control_reg_weight = 1e-1, 1e-3

    freeFlyerQWeight = [0.0] * 3 + [500.0] * 3
    freeFlyerVWeight = [10.0] * 6
    legsQWeight = [0.01] * (rmodel.nv - 6)
    legsWWeights = [1.0] * (rmodel.nv - 6)
    stateWeights = np.array(
        freeFlyerQWeight + legsQWeight + freeFlyerVWeight + legsWWeights
    )

    stateResidual = crocoddyl.ResidualModelState(state, x0, nu)
    stateActivation = crocoddyl.ActivationModelWeightedQuad(stateWeights**2)
    stateReg = crocoddyl.CostModelResidual(state, stateActivation, stateResidual)

    if t == N_ocp:
        costModel.addCost("stateReg", stateReg, state_reg_weight * dt)
    else:
        costModel.addCost("stateReg", stateReg, state_reg_weight)

    if t != N_ocp:
        ctrlResidual = crocoddyl.ResidualModelControl(state, nu)
        ctrlReg = crocoddyl.CostModelResidual(state, ctrlResidual)
        costModel.addCost("ctrlReg", ctrlReg, control_reg_weight)

    # Add COM task
    com_residual = crocoddyl.ResidualModelCoMPosition(state, comDes[t], nu)
    com_activation = crocoddyl.ActivationModelWeightedQuad(np.array([1.0, 1.0, 1.0]))
    com_track = crocoddyl.CostModelResidual(state, com_activation, com_residual)
    if t == N_ocp:
        costModel.addCost("comTrack", com_track, 1e5)
    else:
        costModel.addCost("comTrack", com_track, 1e5)

    constraintModelManager = crocoddyl.ConstraintModelManager(state, actuation.nu)
    if FRICTION_CSTR:
        if t != N_ocp:
            for frame_idx in supportFeetIds:
                name = rmodel.frames[frame_idx].name + "_contact"
                residualFriction = utils_solo12.ResidualFrictionCone(
                    state, name, MU, actuation.nu
                )
                constraintFriction = crocoddyl.ConstraintModelResidual(
                    state, residualFriction, np.array([0.0]), np.array([np.inf])
                )
                constraintModelManager.addConstraint(
                    name + "friction", constraintFriction
                )

    dmodel = crocoddyl.DifferentialActionModelContactFwdDynamics(
        state, actuation, contactModel, costModel, constraintModelManager, 0.0, True
    )
    model = crocoddyl.IntegratedActionModelEuler(dmodel, dt)

    running_models += [model]

# Create shooting problem
ocp = crocoddyl.ShootingProblem(x0, running_models[:-1], running_models[-1])

solver = mim_solvers.SolverCSQP(ocp)
solver.max_qp_iters = 1000
max_iter = 500
solver.setCallbacks([mim_solvers.CallbackVerbose()])
solver.use_filter_line_search = False
solver.termination_tolerance = 1e-4
solver.mu_dynamic = -1  # use Nocedal's L1 merit based on Lagrange multipliers norm
solver.lag_mul_inf_norm_coef = 10.0
solver.eps_abs = 1e-6
solver.eps_rel = 1e-6


xs = [x0] * (solver.problem.T + 1)
us = solver.problem.quasiStatic([x0] * solver.problem.T)

solver.setCallbacks([mim_solvers.CallbackVerbose(), mim_solvers.CallbackLogger()])

solver.solve(xs, us, max_iter)
solution = utils_solo12.get_solution_trajectories(solver, rmodel, rdata, supportFeetIds)

log = solver.getCallbacks()[-1]
crocoddyl.plotOCSolution(solver.xs, solver.us)
mim_solvers.plotConvergence(log.convergence_data)


# Plot solution of the constrained OCP
if PLOT_OCP_SOL:
    # Plot forces
    time_lin = np.linspace(0, T, solver.problem.T)
    fig, axs = plt.subplots(4, 3, constrained_layout=True)
    for i, frame_idx in enumerate(supportFeetIds):
        ct_frame_name = rmodel.frames[frame_idx].name + "_contact"
        forces = np.array(solution[ct_frame_name])
        axs[i, 0].plot(time_lin, forces[:, 0], label="Fx")
        axs[i, 1].plot(time_lin, forces[:, 1], label="Fy")
        axs[i, 2].plot(time_lin, forces[:, 2], label="Fz")
        Fz_lb = (1.0 / MU) * np.sqrt(forces[:, 0] ** 2 + forces[:, 1] ** 2)
        axs[i, 2].plot(time_lin, Fz_lb, "k-.", label="lb")
        axs[i, 0].grid()
        axs[i, 1].grid()
        axs[i, 2].grid()
        axs[i, 0].set_ylabel(ct_frame_name)
    axs[0, 0].legend()
    axs[0, 1].legend()
    axs[0, 2].legend()

    axs[3, 0].set_xlabel(r"$F_x$")
    axs[3, 1].set_xlabel(r"$F_y$")
    axs[3, 2].set_xlabel(r"$F_z$")
    fig.suptitle("Force", fontsize=16)

    comDes = np.array(comDes)
    centroidal_sol = np.array(solution["centroidal"])
    plt.figure()
    plt.plot(comDes[:, 0], comDes[:, 1], "--", label="reference")
    plt.plot(centroidal_sol[:, 0], centroidal_sol[:, 1], label="solution")
    plt.legend()
    plt.xlabel("x")
    plt.ylabel("y")
    plt.title("COM trajectory")
    plt.show()


# Animate solution tof the constrained OCP
if PLAY_OCP_SOL:
    robot = solo12.robot
    viz = pin.visualize.MeshcatVisualizer(
        robot.model, robot.collision_model, robot.visual_model
    )
    try:
        viz.initViewer(open=True)
    except ImportError as err:
        print(err)
        sys.exit(0)
    viz.loadViewerModel()

    arrow1 = utils_solo12.Arrow(
        viz.viewer,
        "force_1",
        location=[0, 0, 0],
        vector=[0, 0, 0.01],
        length_scale=0.05,
    )
    arrow2 = utils_solo12.Arrow(
        viz.viewer,
        "force_2",
        location=[0, 0, 0],
        vector=[0, 0, 0.01],
        length_scale=0.05,
    )
    arrow3 = utils_solo12.Arrow(
        viz.viewer,
        "force_3",
        location=[0, 0, 0],
        vector=[0, 0, 0.01],
        length_scale=0.05,
    )
    arrow4 = utils_solo12.Arrow(
        viz.viewer,
        "force_4",
        location=[0, 0, 0],
        vector=[0, 0, 0.01],
        length_scale=0.05,
    )

    cone1 = utils_solo12.Cone(
        viz.viewer, "friction_cone_1", location=supportFeePos[0], mu=MU
    )
    cone2 = utils_solo12.Cone(
        viz.viewer, "friction_cone_2", location=supportFeePos[1], mu=MU
    )
    cone3 = utils_solo12.Cone(
        viz.viewer, "friction_cone_3", location=supportFeePos[2], mu=MU
    )
    cone4 = utils_solo12.Cone(
        viz.viewer, "friction_cone_4", location=supportFeePos[3], mu=MU
    )

    arrows = [arrow1, arrow2, arrow3, arrow4]
    forces = []

    for i, contactLoc in enumerate(supportFeePos):
        ct_frame_name = rmodel.frames[supportFeetIds[i]].name + "_contact"
        forces.append(np.array(solution[ct_frame_name])[:, :3])
        arrows[i].set_location(contactLoc)

    import time

    viz.display(solution["jointPos"][0])
    time.sleep(2)
    for t in range(N_ocp):
        time.sleep(dt)
        viz.display(solution["jointPos"][t])

        for i in range(len(supportFeePos)):
            arrows[i].anchor_as_vector(supportFeePos[i], forces[i][t])



"""
License: BSD 3-Clause License
Copyright (C) 2024, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.

This file checks that the stagewise QP solver matches the other QP solvers implemented in python
on the solo CoM example (with constraints)
"""

import importlib.util
import os
import pathlib
import time

import crocoddyl
import example_robot_data
import mim_solvers
import numpy as np
import pinocchio
import pinocchio as pin

python_path = pathlib.Path(__file__).absolute().parent.parent.parent / "python"
os.sys.path.insert(1, str(python_path))

from csqp import CSQP  # noqa: E402

HPIPM_PYTHON_FOUND = importlib.util.find_spec("hpipm_python")


class ResidualFrictionCone(crocoddyl.ResidualModelAbstract):
    def __init__(self, state, contact_name, mu, nu):
        crocoddyl.ResidualModelAbstract.__init__(self, state, 1, nu, True, True, True)
        self.mu = mu
        self.contact_name = contact_name

        self.dcone_df = np.zeros((1, 3))
        self.df_dx = np.zeros((3, self.state.ndx))
        self.df_du = np.zeros((3, self.nu))

    def calc(self, data, x, u=None):
        F = data.shared.contacts.contacts[self.contact_name].f.vector[:3]
        data.r[0] = np.array([self.mu * F[2] - np.sqrt(F[0] ** 2 + F[1] ** 2)])

    def calcDiff(self, data, x, u=None):
        F = data.shared.contacts.contacts[self.contact_name].f.vector[:3]

        self.dcone_df[0, 0] = -F[0] / np.sqrt(F[0] ** 2 + F[1] ** 2)
        self.dcone_df[0, 1] = -F[1] / np.sqrt(F[0] ** 2 + F[1] ** 2)
        self.dcone_df[0, 2] = self.mu

        self.df_dx = data.shared.contacts.contacts[self.contact_name].df_dx[:3]
        self.df_du = data.shared.contacts.contacts[self.contact_name].df_du[:3]

        data.Rx = self.dcone_df @ self.df_dx
        data.Ru = self.dcone_df @ self.df_du


pinRef = pin.LOCAL_WORLD_ALIGNED
FRICTION_CSTR = True
MU = 0.8  # friction coefficient

PLOT_OCP_SOL = False
PLAY_OCP_SOL = True

robot_name = "solo12"
ee_frame_names = ["FL_FOOT", "FR_FOOT", "HL_FOOT", "HR_FOOT"]
solo12 = example_robot_data.ROBOTS[robot_name]()
rmodel = solo12.robot.model
rmodel.type = "QUADRUPED"
rmodel.foot_type = "POINT_FOOT"
rdata = rmodel.createData()

# set contact frame_names and_indices
lfFootId = rmodel.getFrameId(ee_frame_names[0])
rfFootId = rmodel.getFrameId(ee_frame_names[1])
lhFootId = rmodel.getFrameId(ee_frame_names[2])
rhFootId = rmodel.getFrameId(ee_frame_names[3])


q0 = np.array(
    [0.0, 0.0, 0.25, 0.0, 0.0, 0.0, 1.0] + 2 * [0.0, 0.8, -1.6] + 2 * [0.0, -0.8, 1.6]
)

x0 = np.concatenate([q0, np.zeros(rmodel.nv)])

pinocchio.forwardKinematics(rmodel, rdata, q0)
pinocchio.updateFramePlacements(rmodel, rdata)
rfFootPos0 = rdata.oMf[rfFootId].translation
rhFootPos0 = rdata.oMf[rhFootId].translation
lfFootPos0 = rdata.oMf[lfFootId].translation
lhFootPos0 = rdata.oMf[lhFootId].translation

comRef = (rfFootPos0 + rhFootPos0 + lfFootPos0 + lhFootPos0) / 4
comRef[2] = pinocchio.centerOfMass(rmodel, rdata, q0)[2].item()

supportFeetIds = [lfFootId, rfFootId, lhFootId, rhFootId]
supportFeePos = [lfFootPos0, rfFootPos0, lhFootPos0, rhFootPos0]


state = crocoddyl.StateMultibody(rmodel)
actuation = crocoddyl.ActuationModelFloatingBase(state)
nu = actuation.nu


comDes = []

N_ocp = 100
dt = 0.02
T = N_ocp * dt
radius = 0.065
for t in range(N_ocp + 1):
    comDes_t = comRef.copy()
    w = (2 * np.pi) * 0.2
    comDes_t[0] += radius * np.sin(w * t * dt)
    comDes_t[1] += radius * (np.cos(w * t * dt) - 1)
    comDes += [comDes_t]

running_models = []
constraintModels = []
for t in range(N_ocp + 1):
    contactModel = crocoddyl.ContactModelMultiple(state, nu)
    costModel = crocoddyl.CostModelSum(state, nu)

    # Add contact
    for frame_idx in supportFeetIds:
        support_contact = crocoddyl.ContactModel3D(
            state,
            frame_idx,
            np.array([0.0, 0.0, 0.0]),
            pinRef,
            nu,
            np.array([0.0, 0.0]),
        )
        contactModel.addContact(
            rmodel.frames[frame_idx].name + "_contact", support_contact
        )

    # Add state/control reg costs

    state_reg_weight, control_reg_weight = 1e-1, 1e-3

    freeFlyerQWeight = [0.0] * 3 + [500.0] * 3
    freeFlyerVWeight = [10.0] * 6
    legsQWeight = [0.01] * (rmodel.nv - 6)
    legsWWeights = [1.0] * (rmodel.nv - 6)
    stateWeights = np.array(
        freeFlyerQWeight + legsQWeight + freeFlyerVWeight + legsWWeights
    )

    stateResidual = crocoddyl.ResidualModelState(state, x0, nu)
    stateActivation = crocoddyl.ActivationModelWeightedQuad(stateWeights**2)
    stateReg = crocoddyl.CostModelResidual(state, stateActivation, stateResidual)

    if t == N_ocp:
        costModel.addCost("stateReg", stateReg, state_reg_weight * dt)
    else:
        costModel.addCost("stateReg", stateReg, state_reg_weight)

    if t != N_ocp:
        ctrlResidual = crocoddyl.ResidualModelControl(state, nu)
        ctrlReg = crocoddyl.CostModelResidual(state, ctrlResidual)
        costModel.addCost("ctrlReg", ctrlReg, control_reg_weight)

        FORCE_COST = True
        if FORCE_COST:
            for frame_idx in supportFeetIds:
                fref = pin.Force.Zero()  # np.zeros(3)
                # fref[2] = 1
                ForceResidual = crocoddyl.ResidualModelContactForce(
                    state, frame_idx, fref, 3, nu
                )
                ForceRes = crocoddyl.CostModelResidual(state, ForceResidual)
                costModel.addCost("ForceRes" + str(frame_idx), ForceRes, 1.0)

    # Add COM task
    com_residual = crocoddyl.ResidualModelCoMPosition(state, comDes[t], nu)
    com_activation = crocoddyl.ActivationModelWeightedQuad(np.array([1.0, 1.0, 1.0]))
    com_track = crocoddyl.CostModelResidual(state, com_activation, com_residual)
    if t == N_ocp:
        costModel.addCost("comTrack", com_track, 1e5)
    else:
        costModel.addCost("comTrack", com_track, 1e5)

    constraintModelManager = crocoddyl.ConstraintModelManager(state, actuation.nu)
    if FRICTION_CSTR:
        if t != N_ocp:
            for frame_idx in supportFeetIds:
                name = rmodel.frames[frame_idx].name + "_contact"
                residualFriction = ResidualFrictionCone(state, name, MU, actuation.nu)
                constraintFriction = crocoddyl.ConstraintModelResidual(
                    state, residualFriction, np.array([0.0]), np.array([np.inf])
                )
                constraintModelManager.addConstraint(
                    name + "friction", constraintFriction
                )

    dmodel = crocoddyl.DifferentialActionModelContactFwdDynamics(
        state, actuation, contactModel, costModel, constraintModelManager, 0.0, True
    )
    model = crocoddyl.IntegratedActionModelEuler(dmodel, dt)

    running_models += [model]

# Create shooting problem
problem = crocoddyl.ShootingProblem(x0, running_models[:-1], running_models[-1])


# # # # # # # # # # # # #
###     SOLVE OCP     ###
# # # # # # # # # # # # #

# Define warm start
xs_init = [x0] * (N_ocp + 1)
us_init = problem.quasiStatic([problem.x0] * problem.T)

# Define solver
ddp0 = mim_solvers.SolverCSQP(problem)  # CSQP(problem, "StagewiseQP")
ddp1 = CSQP(problem, "ProxQP")
ddp2 = CSQP(problem, "OSQP")
if HPIPM_PYTHON_FOUND is not None:
    ddp4 = CSQP(problem, "HPIPM_DENSE")
    ddp5 = CSQP(problem, "HPIPM_OCP")

ddp0.with_callbacks = False
ddp1.with_callbacks = False
ddp2.with_callbacks = False
if HPIPM_PYTHON_FOUND is not None:
    ddp4.with_callbacks = False
    ddp5.with_callbacks = False

max_qp_iters = 10000
ddp0.max_qp_iters = max_qp_iters
ddp1.max_qp_iters = max_qp_iters
ddp2.max_qp_iters = max_qp_iters
if HPIPM_PYTHON_FOUND is not None:
    ddp4.max_qp_iters = max_qp_iters
    ddp5.max_qp_iters = max_qp_iters

eps_abs = 1e-10
eps_rel = 0.0
ddp0.eps_abs = eps_abs
ddp0.eps_rel = eps_rel
ddp1.eps_abs = eps_abs
ddp1.eps_rel = eps_rel
ddp2.eps_abs = eps_abs
ddp2.eps_rel = eps_rel
if HPIPM_PYTHON_FOUND is not None:
    ddp4.eps_abs = eps_abs
    ddp4.eps_rel = eps_rel
    ddp5.eps_abs = eps_abs
    ddp5.eps_rel = eps_rel


ddp0.equality_qp_initial_guess = False
ddp1.equality_qp_initial_guess = False
ddp2.equality_qp_initial_guess = False
if HPIPM_PYTHON_FOUND is not None:
    ddp4.equality_qp_initial_guess = False
    ddp5.equality_qp_initial_guess = False

ddp0.update_rho_with_heuristic = True


# Stagewise QP
print("\n ------ STAGEWISE (computeDirection)------ ")
converged = ddp0.solve(xs_init, us_init, 0)
t0 = time.time()
ddp0.computeDirection(True)
print("Stagewise : ", time.time() - t0)
print("------------------------ \n")

# ProxQP
print("\n ------ ProxQP ------ ")
converged = ddp1.solve(xs_init, us_init, 1)
print("------------------------ \n")


# OSQP
print("\n ------ OSQP ------ ")
converged = ddp2.solve(xs_init, us_init, 1)
print("------------------------ \n")

if HPIPM_PYTHON_FOUND is not None:
    # HPIPM
    print("\n ------ HPIPM DENSE ------ ")
    converged = ddp4.solve(xs_init, us_init, 1)
    print("------------------------ \n")

    # HPIPM
    print("\n ------ HPIPM OCP ------ ")
    converged = ddp5.solve(xs_init, us_init, 1)
    print("------------------------ \n")

#  iterations
print("Stagewise iter = ", int(ddp0.qp_iters))
print("ProxQP iter = ", int(ddp1.qp_iters))
print("OSQP iter      = ", ddp2.qp_iters)
if HPIPM_PYTHON_FOUND is not None:
    print("HPIPM DENSE iter     = ", ddp4.qp_iters)
    print("HPIPM OCP iter     = ", ddp5.qp_iters)

# Check that QP solutions are the same
TOL = 1e-2
assert np.linalg.norm(np.array(ddp0.dx_tilde.tolist()) - np.array(ddp1.dx)) < TOL
assert np.linalg.norm(np.array(ddp0.dx_tilde.tolist()) - np.array(ddp2.dx)) < TOL
if HPIPM_PYTHON_FOUND is not None:
    assert np.linalg.norm(np.array(ddp0.dx_tilde.tolist()) - np.array(ddp4.dx)) < TOL
    assert np.linalg.norm(np.array(ddp0.dx_tilde.tolist()) - np.array(ddp5.dx)) < TOL
    assert np.linalg.norm(np.array(ddp4.dx) - np.array(ddp5.dx)) < TOL
assert np.linalg.norm(np.array(ddp0.du_tilde.tolist()) - np.array(ddp1.du)) < TOL
assert np.linalg.norm(np.array(ddp0.du_tilde.tolist()) - np.array(ddp2.du)) < TOL
if HPIPM_PYTHON_FOUND is not None:
    assert np.linalg.norm(np.array(ddp0.du_tilde.tolist()) - np.array(ddp4.du)) < TOL
    assert np.linalg.norm(np.array(ddp0.du_tilde.tolist()) - np.array(ddp5.du)) < TOL
    assert np.linalg.norm(np.array(ddp4.du) - np.array(ddp5.du)) < TOL



"""__init__
License: BSD 3-Clause License
Copyright (C) 2023, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.
"""

from .mim_solvers_pywrap import *  # noqa
import copy
import numpy as np


class CallbackLogger(CallbackAbstract):  # noqa: F405
    def __init__(self, verbose=False):
        super().__init__()
        self.convergence_data = {}

    def __call__(self, solver, solver_type):
        def safe_append(key, value):
            if key not in self.convergence_data:
                self.convergence_data[key] = []
            self.convergence_data[key].append(value)

        if solver_type == "CSQP":
            safe_append("us", copy.copy((solver.us)))
            safe_append("xs", copy.copy((solver.xs)))
            safe_append("fs", copy.copy((solver.fs)))
            safe_append("iter", solver.iter)
            safe_append("cost", solver.cost)
            safe_append("merit", solver.merit)
            safe_append("preg", solver.preg)
            safe_append("dreg", solver.dreg)
            safe_append("step", solver.stepLength)
            safe_append("gap_norm", solver.gap_norm)
            safe_append("constraint_norm", solver.constraint_norm)
            safe_append("qp_iter", solver.qp_iters)
            safe_append("KKT", solver.KKT)
        elif solver_type == "SQP":
            safe_append("us", copy.copy((solver.us)))
            safe_append("xs", copy.copy((solver.xs)))
            safe_append("fs", copy.copy((solver.fs)))
            safe_append("iter", solver.iter)
            safe_append("cost", solver.cost)
            safe_append("merit", solver.merit)
            safe_append("preg", solver.preg)
            safe_append("dreg", solver.dreg)
            safe_append("step", solver.stepLength)
            safe_append("KKT", solver.KKT)
        else:
            raise NotImplementedError("CallbackLogger is implemented for CSQP and SQP.")


def plotConvergence(data, show=True):
    import matplotlib.pyplot as plt

    axis = 0
    for i, (key, values) in enumerate(data.items()):
        if len(np.asarray(values, dtype="object").shape) == 1:
            axis += 1
    fig, axs = plt.subplots(
        axis - 2, 1, sharex="col", figsize=(55, 25.5)
    )  # exclude dreg and preg

    i = 0
    for key, values in data.items():
        if len(np.asarray(values, dtype="object").shape) == 1:
            if key == "dreg" or key == "preg":
                continue

            axs[i].plot(values)
            axs[i].set_title(key)
            if key == "qp_iter":
                axs[i].text(
                    0.9,
                    0.9,
                    f"Total qp_iters: {sum(values)}",
                    transform=axs[i].transAxes,
                    ha="right",
                )
                axs[i].text(
                    0.9,
                    0.8,
                    f"Total sqp_iters: {len(values)}",
                    transform=axs[i].transAxes,
                    ha="right",
                )

            if key == "KKT":
                axs[i].set_yscale("log")
            i += 1
    plt.tight_layout()

    if show:
        plt.show()

    return fig
